import os
import logging
import tempfile
import pandas as pd
import numpy as np
import subprocess as sp
from collections import defaultdict
from operator import itemgetter
from wgd.peak import formatv2
from wgd.viz import apply_filters

gff_header = ["gene", "scaffold", "start", "orientation"] 

# we construct first a table with each row a gene containing it's family,
# scaffold, location and orientation
def make_gene_table(gffs, families, feature, attribute, additionalgffinfo):
    """
    Construct a table from a bunch of gff files and gene families such that all
    information for synteny and co-linearity related analyses is in one place.
    
    :param gffs: a list of GFF *file names*
    :param families: a pandas data frame with the gene families
    :param feature: feature tag for the GFF files
    :param attribute: attribute tag for parsing out gene IDs from last column
    
    Note: currently we assume the same attribute/feature is required for all
    GFF files, and the burden is on the user to make sure this is the case. We
    may wish, for flexibility, to allow a list of feature/attribute arguments
    in the future, one for each GFF file.
    """
    if not (additionalgffinfo is None) and len(additionalgffinfo) > 0:
        # ignore the info in the original feature and attribute
        features, attributes = [i.split(";")[0].strip() for i in additionalgffinfo], [i.split(";")[1].strip() for i in additionalgffinfo]
        if len(features) != len(attributes):
            logging.error("Please give the additional feature and attribute info of gff in the format of (feature;attribute)")
            exit(1)
        if len(features) != len(gffs):
            logging.error("Please give the same number and order of gff files and its associated additional feature and attribute info")
            exit(1)
        gfftables = [gff2table(gff, f, a) for gff,f,a in zip(gffs,features,attributes)]
    else:
        gfftables = [gff2table(gff, feature, attribute) for gff in gffs]
    familytable = gene2family(families)
    df = pd.concat(gfftables)
    df = familytable.join(df)
    return df

def getattr(s, attribute):
    for x in s.split(";"):
        y = x.split("=")
        if y[0] == attribute:
            return y[1].strip()
    return ""

def gff2table(gff, feature, attribute):
    """
    Read a GFF file to a pandas data frame, from a filename.
    """
    rows = []
    with open(gff, "r") as f:
        for l in f.readlines():
            if l.startswith("#") or l.strip("\n").strip()=='':
                continue
            x = l.strip("\n").strip("\t").split("\t")
            #Note here the empty lines from input will make error
            if x[2] == feature:
                a = getattr(x[-1], attribute)
                rows.append({"gene": a, "scaffold": x[0], "start": int(x[3]), "or": x[6]})
    df = pd.DataFrame.from_dict(rows).set_index("gene")
    return df

def gene2family(families):
    """
    Get a gene to family ID mapping from a gene families data frame in
    OrthoFinder format.
    """
    rows = []
    for fam in families.index:
        for sp in families.columns:
            x = families.loc[fam, sp]
            if type(x) != str:
                continue
            for gene in x.split(", "):
                rows.append({"gene": gene.strip(), 
                    "species": sp, "family": fam})
    df = pd.DataFrame.from_dict(rows).set_index("gene")
    return df

def configure_adhore(table, outdir, **kwargs):
    """
    Write out all required files for running I-ADHoRe.
    """
    if not os.path.exists(outdir):
        os.mkdir(outdir)
    famfile = os.path.join(outdir, "families.tsv")
    logging.info("Writing families file")
    write_families(famfile, table)
    logging.info("Writing gene lists")
    genelists = write_genelists(outdir, table)
    logging.info("Writing config file")
    config_file = os.path.join(outdir, "iadhore.conf")
    out_path = os.path.abspath(os.path.join(outdir, "iadhore-out"))
    write_config_adhore(config_file, out_path, genelists, os.path.abspath(famfile), **kwargs)
    return config_file, out_path

def write_families(fname, table):
    table["family"].to_csv(fname, sep="\t", header=None)

def write_genelists(outdir, table):
    genomes = {}
    for sp, df in table.groupby("species"):
        gdir = os.path.join(outdir, sp + ".lists")
        if not os.path.exists(gdir):
            os.mkdir(gdir)
        lists = {}
        for scaffold, sdf in df.groupby("scaffold"):
            if len(sdf.index) <= 2: continue
            fname = os.path.join(gdir, scaffold)
            with open(fname, "w") as o:
                if len(list(sdf.sort_values(by=["start"]).index)) != len(set(sdf.sort_values(by=["start"]).index)):
                    logging.error("There are duplicated gene IDs for given feature and attribute")
                    exit(1)
                for g in sdf.sort_values(by=["start"]).index:
                    o.write(g + sdf.loc[g,"or"] + "\n")
            lists[scaffold] = os.path.abspath(fname)
        genomes[sp] = lists
    return genomes

def write_config_adhore(
        config_file, output_path, genelists, families, gap_size=30,
        cluster_gap=35, q_value=0.75, prob_cutoff=0.01, anchor_points=3,
        alignment_method='gg2', level_2_only='false', table_type='family',
        multiple_hypothesis_correction='FDR', visualizeGHM='false',
        visualizeAlignment='false', number_of_threads=4):
    """
    Write out the config file for I-ADHoRe. See I-ADHoRe manual for information
    on parameter settings.

    :param gene_lists: directory with gene lists per chromosome
    :param families: file with gene to family mapping
    :param config_file_name: name for the config file
    :param output_path: output path name
    :param gap_size: see I-ADHoRe 3.0 documentation
    :param cluster_gap: see I-ADHoRe 3.0 documentation
    :param q_value: see I-ADHoRe 3.0 documentation
    :param prob_cutoff: see I-ADHoRe 3.0 documentation
    :param anchor_points: see I-ADHoRe 3.0 documentation
    :param alignment_method: see I-ADHoRe 3.0 documentation
    :param level_2_only: see I-ADHoRe 3.0 documentation
    :param table_type: see I-ADHoRe 3.0 documentation
    :param multiple_hypothesis_correction: see I-ADHoRe 3.0 documentation
    :param visualizeGHM: see I-ADHoRe 3.0 documentation
    :param visualizeAlignment: see I-ADHoRe 3.0 documentation
    :param number_of_threads: see I-ADHoRe 3.0 documentation
    :return: configuration file see I-ADHoRe 3.0 documentation
    """
    with open(config_file, 'w') as o:
        for k,v in genelists.items():
            o.write('genome= {}\n'.format(k))
            for scaffold, fname in v.items():
                o.write("{} {}\n".format(scaffold, fname))
            o.write("\n")

        o.write('blast_table= {}\n'.format(families))
        o.write('output_path= {}\n'.format(output_path))
        o.write('gap_size= {}\n'.format(gap_size))
        o.write('q_value= {}\n'.format(q_value))
        o.write('cluster_gap= {}\n'.format(cluster_gap))
        o.write('prob_cutoff= {}\n'.format(prob_cutoff))
        o.write('anchor_points= {}\n'.format(anchor_points))
        o.write('alignment_method= {}\n'.format(alignment_method))
        o.write('level_2_only= {}\n'.format(level_2_only))
        o.write('table_type= {}\n'.format(table_type))
        o.write('multiple_hypothesis_correction= {}\n'.format(
                multiple_hypothesis_correction))
        o.write('visualizeGHM= {}\n'.format(visualizeGHM))
        o.write('visualizeAlignment= {}\n'.format(visualizeAlignment))
        o.write('number_of_threads= {}\n'.format(number_of_threads))
    return os.path.abspath(output_path)

def run_adhore(config_file):
    """
    Run I-ADHoRe for a given config file

    :param config_file: path to I-ADHoRe configuration file
    """
    completed = sp.run(['i-adhore', config_file], stderr=sp.PIPE, stdout=sp.PIPE)
    logging.warning(completed.stderr.decode('utf-8'))
    logging.info(completed.stdout.decode('utf-8'))
    return

def get_anchors(out_path,userdf=None):
    if userdf!=None: anchors = pd.read_csv(userdf, sep="\t", index_col=0)
    else: anchors = pd.read_csv(os.path.join(out_path, "anchorpoints.txt"), sep="\t", index_col=0)
    if len(anchors) == 0:
        return None, None
    orig_anchors = anchors.copy()
    anchors["pair"] = anchors[["gene_x", "gene_y"]].apply(lambda x: "__".join(sorted([x[0], x[1]])), axis=1)
    df = anchors[["pair", "multiplicon"]].drop_duplicates("pair").set_index("pair")
    #anchors["pair_reverse"] = anchors[["gene_x", "gene_y"]].apply(lambda x: "__".join(sorted([x[1], x[0]])), axis=1)
    #df_reverse = anchors[["pair_reverse", "multiplicon"]].drop_duplicates("pair_reverse").set_index("pair_reverse")
    #df_reverse.index.set_names('pair',inplace=True)
    # there are duplicates, due to anchors being in multiple multiplicons
    return df, orig_anchors

def get_multi(out_path,userdf2=None):
    if userdf2!=None: multi = pd.read_csv(userdf2, sep="\t", index_col=None,header = 0)
    else: multi = pd.read_csv(os.path.join(out_path, "multiplicons.txt"), sep="\t", index_col=None,header = 0)
    return multi

def get_anchor_ksd(ks_distribution, anchors):
    return ks_distribution.join(anchors).dropna()

def get_segments_profile(multi,keepredun,out_path,userdf3=None):
    if userdf3!=None: segs = pd.read_csv(userdf3, sep="\t", index_col=0)
    else: segs = pd.read_csv(os.path.join(out_path, "segments.txt"), sep="\t", index_col=0)
    if not keepredun:
        I3 = []
        Mul_to_rm = list(multi[multi['is_redundant']==-1].loc[:,'id'])
        Segs_to_rm = segs.loc[segs['multiplicon'].isin(Mul_to_rm),:]
        for i in Segs_to_rm.index: I3.append(i)
        segs = segs.drop(I3)
    segs["segment"] = segs.index
    return segs

def get_chrom_gene(table,outdir):
    df = table.reset_index()
    gdf = list(df.groupby("species"))
    scafs_genes,ordered_dfs,gene_orders = {},{},{}
    for sp, df in gdf:
        scafs_genes[sp] = {}
        for i in set(df['scaffold']): scafs_genes[sp][i] = []
        tmp = list(df.groupby(['scaffold'])[['gene','start']])
        for i in tmp:
            df_tmp = i[1].sort_values(by=['start'])
            for indice,j in enumerate(df_tmp['gene']):
                scafs_genes[sp][i[0]].append(j)
                if gene_orders.get(j) == None:
                    gene_orders[j] = indice+1 # since we want the gene starting from 1 instead of 0
        max_length = max(len(v) for v in scafs_genes[sp].values())
        for key in scafs_genes[sp].keys():
            if len(scafs_genes[sp][key]) < max_length:
                scafs_genes[sp][key].extend([None] * (max_length - len(scafs_genes[sp][key])))
        df_output = pd.DataFrame(scafs_genes[sp])
        ordered_df = df_output[sorted(df_output.columns,key=lambda y: len(df_output[y].dropna()),reverse=True)]
        fname = os.path.join(outdir, "{}_gene_order_perchrom.tsv".format(sp))
        ordered_df.insert(0, "Coordinates", np.arange(1,max_length+1))
        ordered_df.fillna('').to_csv(fname,header=True,index=False,sep='\t')
        ordered_dfs[sp] = ordered_df
    return ordered_dfs, gene_orders

def getunitMP(MP,gene_scaf):
    MP['scaffold_x'] = MP['gene_x'].apply(lambda x:gene_scaf[x])
    MP['scaffold_y'] = MP['gene_y'].apply(lambda x:gene_scaf[x])
    units = []
    for indice, sx, sy in zip(range(len(MP)),MP['scaffold_x'],MP['scaffold_y']):
        if indice == 0:
            unit = 0
            units.append(unit)
            sxy = (sx,sy)
            continue
        if (sx,sy) == sxy:
            units.append(unit)
            continue
        unit = unit + 1
        sxy = (sx,sy)
        units.append(unit)
    MP['unit'] = units
    return MP

def getorien(MP):
    orien = []
    for unit, df in list(MP.groupby('unit')):
        if len(df) == 1:
            orien.append("x")
            continue
        posix,negax,posiy,negay = 0,0,0,0
        for indice, corx, cory in zip(range(len(df)),df['coordinate_x'],df['coordinate_y']):
            if indice == 0:
                cx,cy = corx,cory
                continue
            if corx - cx > 0:
                posix = posix + 1
            if corx - cx < 0:
                negax = negax + 1
            if cory - cy > 0:
                posiy = posiy + 1
            if cory - cy < 0:
                negay = negay + 1
            cx,cy = corx,cory
        Judge = (posix - negax) * (posiy - negay)
        for i in range(len(df)):
            if Judge > 0:
                orien.append("+")
            if Judge < 0:
                orien.append("-")
            if Judge == 0:
                orien.append("x")
    MP['orientation'] = orien
    return MP

def get_mp_geneorder(gene_orders,out_path,outdir,table,userdf4=None):
    gene_scaf = {g:s for g,s in zip(table.index,table['scaffold'])}
    if userdf4 != None: MP = pd.read_csv(userdf4, sep="\t", index_col=0)
    else: MP = pd.read_csv(os.path.join(out_path, "multiplicon_pairs.txt"), sep="\t", index_col=0)
    MP = MP.rename(columns={"gene_y": "code"})
    MP = MP.rename(columns={"Unnamed: 2": "gene_x", "gene_x": "gene_y"})
    MP['coordinate_x'] = MP['gene_x'].apply(lambda x:gene_orders[x])
    MP['coordinate_y'] = MP['gene_y'].apply(lambda x:gene_orders[x])
    MP = getunitMP(MP,gene_scaf)
    MP = getorien(MP)
    fname = os.path.join(outdir, "multiplicon_pairs_coordinates_unit_orien.tsv")
    MP.to_csv(fname,header=True,index=True,sep='\t')
    return MP

def transformunit(segs,ordered_genes_perchrom_allsp,outdir):
    gene_order_dict_allsp = {}
    seg = segs.copy()
    splist = set(seg['genome'])
    for sp in splist:
        gene_order_dict_allsp[sp] = {}
        ordered_genes_perchrom = ordered_genes_perchrom_allsp[sp]
        ordered_genes_perchrom = ordered_genes_perchrom.set_index('Coordinates')
        for i in ordered_genes_perchrom.index:
            for scaf in ordered_genes_perchrom.columns:
                genename = ordered_genes_perchrom.loc[i,scaf]
                gene_order_dict_allsp[sp][genename] = i
    first_coordinate,last_coordinate = [],[]
    for i,j in zip(seg['genome'],seg['first']): first_coordinate.append(gene_order_dict_allsp[i][j])
    for i,j in zip(seg['genome'],seg['last']): last_coordinate.append(gene_order_dict_allsp[i][j])
    seg['first_coordinate'],seg['last_coordinate'] = first_coordinate,last_coordinate
    fname = os.path.join(outdir, "segments_coordinates.tsv")
    seg.to_csv(fname,header=True,index=True,sep='\t')
    return seg, gene_order_dict_allsp

def getsegks(segs_gene_unit,ks_distribution,ordered_genes_perchrom_allsp):
    ksdb_df = pd.read_csv(ks_distribution,header=0,index_col=0,sep='\t')
    ksdb_df = formatv2(ksdb_df)
    df = apply_filters(ksdb_df, [("dS", 0., 5.)])
    for indice in segs_gene_unit.index:
        sp,gl = segs_gene_unit[indice,'genome'],segs_gene_unit[indice,'list']
        coor_s, coor_l = segs_gene_unit[indice,'first_coordinate'],segs_gene_unit[indice,'last_coordinate']
        df_sp = ordered_genes_perchrom_allsp[sp]


def annotatelist(anchorpoints,table):
    ap_with_list = anchorpoints.copy()
    gene_list = {gene:li for gene,li in zip(table.index,table['scaffold'])}
    ap_with_list['list_x'] = ap_with_list['gene_x'].apply(lambda x:gene_list[x])
    ap_with_list['list_y'] = ap_with_list['gene_y'].apply(lambda x:gene_list[x])
    return ap_with_list

def annotatesegy(ap_with_list,segs):
    tmp = list(ap_with_list.groupby['multiplicon'])
    unique_segid = []
    for mlt, df in tmp:
        df = df.sort_values(by=['coord_y'])
        gl = list(df['list_y'])[0]
        starty,endy = list(df['gene_y'])[0], list(df['gene_y'])[-1] # ap should be inside the segment for sure
        seg_tmp = segs.loc[(segs['multiplicon']==mlt) & (segs['list']==gl),:]
        if len(seg_tmp) == 1:
            unique_segid.append(seg_tmp['segment'])
            continue

def getsegid(starty,endy,segs,gene_list,mlt,gene_orders):
    gls,gle = gene_list[starty],gene_list[endy]
    assert gls == gle
    tmp = segs.loc[(segs['multiplicon']==mlt) & (segs['list']==gls),:]
    if len(tmp) == 1:
        return list(tmp['segment'])[0]
    else:
        for segid,first,last in zip(tmp['segment'],tmp['first'],tmp['last']):
            f_coor,l_coor = gene_orders[first], gene_orders[last]
            if gene_orders[starty] >= f_coor and gene_orders[endy] <= l_coor:
                return segid

def getsegidx(df,segs,gene_list,mlt,gene_orders):
    df_tmp = df.copy()
    df_tmp['list_x'] = df_tmp['gene_x'].apply(lambda x:gene_list[x])
    df_tmp = df_tmp.sort_values(by=['list_x','coord_x'])

def annotatelist_full(anchorpoints,table,segs,gene_orders):
    ap_with_list = anchorpoints.copy()
    list_x,list_y = [],[]
    tmp = list(ap_with_list.groupby('multiplicon'))
    gene_list = {gene:li for gene,li in zip(table.index,table['scaffold'])}
    for mlt,df in tmp:
        df = df.sort_values(by=['coord_y'])
        starty,endy = list(df['gene_y'])[0], list(df['gene_y'])[-1]
        belonged_segidy = getsegid(starty,endy,segs,gene_list,mlt,gene_orders)
        for i in range(df.shape[0]): list_y.append(belonged_segidy)
        belonged_segidx = getsegidx(df,segs,gene_list,mlt,gene_orders)

def get_segmentpair_order(anchorpoints,segs,table,gene_orders):
    ap_with_list = annotatelist_full(anchorpoints,table,segs,gene_orders)
    ap_with_list = annotatelist(anchorpoints,table)
    ap_with_list_addsegy = annotatesegy(ap_with_list,segs)

def getmltorder(anchorpoints,multi,gene_orders):
    ap_order_permlt = {i:'' for i in set(anchorpoints['multiplicon'])}
    tmp = list(anchorpoints.groupby('multiplicon'))
    for mlt, df in tmp:
        genexs,geneys = list(df['gene_x'])[:3],list(df['gene_y'])[:3]
        genexs_coor, geneys_coor = [gene_orders[i] for i in genexs], [gene_orders[i] for i in geneys]
        consistent_order = []
        for i in range(2):
            tmpx = genexs_coor[i+1]-genexs_coor[i]
            tmpy = geneys_coor[i+1]-geneys_coor[i]
            if tmpx > 0: pos_neg = "+"
            else: pos_neg = "-"
            consistent_order.append(tmpx*tmpy)
        for i in consistent_order:
            assert i > 0
        ap_order_permlt[mlt] = pos_neg
    return ap_order_permlt



import pytest
import os
import logging
import shutil
from wgd.core import SequenceData, get_gene_families, read_gene_families
from wgd.core import KsDistributionBuilder

# some config: set logging level, and get directory
logging.basicConfig(level=logging.ERROR)
thisdir = os.path.dirname(os.path.abspath(__file__))


# Test the internals
class TestCore:
    @pytest.fixture
    def data(self):
        datadir = os.path.join(thisdir, "data")
        s1 = os.path.join(datadir, "ugi1000.fasta")
        s2 = os.path.join(datadir, "egu1000.fasta")
        return (s1, s2)

    def test_seqio(self, tmpdir, data):
        s1, s2 = data
        logging.info("Testing `SequenceData` (IO)")
        kwargs = {"out_path": tmpdir, "tmp_path": tmpdir}
        d0 = SequenceData(s1, cds=True, to_stop=True, **kwargs)
        d1 = SequenceData(s1, cds=False, to_stop=False, **kwargs)
        d2 = SequenceData(s2, cds=False, to_stop=True, **kwargs)
        assert len(d0.cds_seqs) == 224  # Should only read proper CDS
        assert len(d1.cds_seqs) == 500  # Should read full file
        assert len(d2.pro_seqs) == 500  # Proper translation

    def test_paranome(self, data, tmpdir):
        logging.info("Testing paranome (requires diamond+mcl)")
        s1, s2 = data
        d = SequenceData(s1, out_path=tmpdir, tmp_path=tmpdir, cds=False, cscore=None, to_stop=False)
        d.get_paranome()
        # MCL is not deterministic, but the number of families is somewhere around 35
        assert 30 < len(d.mcl) < 40  

        # round tripping for families?
        # write to file/read from file
        families = d.write_paranome(False,singletons=False)
        fams = read_gene_families(families)
        fams = get_gene_families(d, fams)
        assert len(d.mcl) == len(fams)  # Round tripped

        # construct gene families directly from `self.mcl`
        #fams2 = get_gene_families(d, d.mcl.values(), rename=False)
        #assert len(fams2) == len(fams)

    def test_rbh(self, data, tmpdir):
        logging.info("Testing RBH orthologs (requires diamond)")
        s1, s2 = data
        kwargs = {"out_path": tmpdir, "tmp_path": tmpdir, "cds":False, "to_stop":False}
        d1 = SequenceData(s1, **kwargs)
        d2 = SequenceData(s2, **kwargs)
        d1.get_rbh_orthologs(d2,None,False)
        df = d1.rbh[d2.prefix]
        # test below fails with PAML 4.9i (works with 4.9j)
        # test below fails with diamond 0.9.18 (works with 2.0.5)
        assert len(df.index) == 16  # Got all RBHs 
        assert len(df.columns) == 12  # Right No. columns

    def test_ksd(self, data, tmpdir):
        s1, s2 = data
        d = SequenceData(s1, out_path=tmpdir, tmp_path=tmpdir, cds=True, to_stop=True)
        d.get_paranome()

        # TODO: split in subtests (functions)
        # alignment without gap-stripping
        families = d.write_paranome(False)
        fams = read_gene_families(families)
        gfams = get_gene_families(d, fams, prequal=False, strip_gaps=False)
        f = gfams[0]
        f.align()
        cds_len = f.cds_aln.get_alignment_length()
        pro_len = f.pro_aln.get_alignment_length()
        # test below fails with PAML 4.9i (works with 4.9j)
        # test below fails with diamond 0.9.18 (works with 2.0.5)
        assert cds_len == 558  # Alignment working?
        assert cds_len == 3*pro_len  # "AA -> codon fine?

        # alignment with gap-stripping
        gfams = get_gene_families(d, fams, prequal=False, strip_gaps=True)
        f = gfams[0]
        f.align()
        cds_len = f.cds_aln.get_alignment_length()
        pro_len = f.pro_aln.get_alignment_length()
        gaps = 0
        for i in range(pro_len):
            if "-" in f.pro_aln[:,i]:
                gaps += 1
        assert cds_len == 3*(pro_len - gaps)

        # run Ks distribution construction
        ksdb = KsDistributionBuilder(gfams,d)
        ksdb.get_distribution()
        assert len(ksdb.df.index.unique()) == len(ksdb.df.index)
        assert pytest.approx(25., 1.) == ksdb.df["dS"].mean()
        assert pytest.approx(90., 1.) == ksdb.df["S"].mean()
        
        # pairwise mode
        gfams = get_gene_families(d, fams, pairwise=True,
                prequal=False, strip_gaps=False)
        ksdb = KsDistributionBuilder(gfams,d)
        ksdb.get_distribution()
        assert len(ksdb.df.index.unique()) == len(ksdb.df.index)
        assert pytest.approx(25., 1.) == ksdb.df["dS"].mean()
        assert pytest.approx(90., 1.) == ksdb.df["S"].mean()


# TODO: test CLI



# TODO: clean up, currently obsolete, but we may wish to recycle stuff...
import os
import logging
import re
import warnings
import random
import numpy as np
import json
import subprocess
from progressbar import ProgressBar
from numpy import mean, std
from scipy.spatial.distance import cdist

def formatv2(ksdf):
    if "Ks" in ksdf.columns: ksdf = ksdf.rename(columns={"Ks":"dS"})
    if "Ka" in ksdf.columns: ksdf = ksdf.rename(columns={"Ka":"dN"})
    if "Omega" in ksdf.columns: ksdf = ksdf.rename(columns={"Omega":"dN/dS"})
    if "Family" in ksdf.columns: ksdf = ksdf.rename(columns={"Family":"family"})
    if "Node" in ksdf.columns: ksdf = ksdf.rename(columns={"Node":"node"})
    if "AlignmentIdentity" in ksdf.columns: ksdf = ksdf.rename(columns={"AlignmentIdentity":"alignmentidentity"})
    if "AlignmentLength" in ksdf.columns: ksdf = ksdf.rename(columns={"AlignmentLength":"alignmentlength"})
    if "AlignmentCoverage" in ksdf.columns: ksdf = ksdf.rename(columns={"AlignmentCoverage":"alignmentcoverage"})
    if "Paralog1" in ksdf.columns: ksdf = ksdf.rename(columns={"Paralog1":"gene1"})
    if "Paralog2" in ksdf.columns: ksdf = ksdf.rename(columns={"Paralog2":"gene2"})
    if "WeightOutliersIncluded" in ksdf.columns: ksdf = ksdf.drop(columns=["WeightOutliersIncluded"])
    if "WeightOutliersExcluded" in ksdf.columns: ksdf = ksdf.drop(columns=["WeightOutliersExcluded"])
    if "weightoutlierexcluded" not in ksdf.columns: weight_inc = get_outlierincluded(ksdf)
    if "weightoutlierincluded" not in ksdf.columns:
        weight_exc = get_outlierexcluded(ksdf,cutoff = 5)
        ksdf = ksdf.join(weight_inc).join(weight_exc)
    if "node_averaged_dS_outlierincluded" not in ksdf.columns:
        node_averaged_dS_inc = get_nodeaverged_dS_outlierincluded(ksdf)
    if "node_averaged_dS_outlierexcluded" not in ksdf.columns:
        node_averaged_dS_exc = get_nodeaverged_dS_outlierexcluded(ksdf,cutoff = 5)
        ksdf = ksdf.reset_index().merge(node_averaged_dS_inc,on = ['family', 'node'])
        ksdf = ksdf.merge(node_averaged_dS_exc,on = ['family', 'node'],how = 'left')
    if "index" in ksdf.columns: ksdf = ksdf.set_index('index')
    ksdf.index.name = 'pair'
    return ksdf

def get_outlierexcluded(df,cutoff = 5):
    df = df[df['dS']<cutoff]
    weight_exc = 1/df.groupby(['family', 'node'])['dS'].transform('count')
    weight_exc = weight_exc.to_frame(name='weightoutlierexcluded')
    return weight_exc

def get_outlierincluded(df):
    weight_inc = 1/df.groupby(['family', 'node'])['dS'].transform('count')
    weight_inc = weight_inc.to_frame(name='weightoutlierincluded')
    return weight_inc

def get_nodeaverged_dS_outlierincluded(df):
    node_averaged_dS_inc = df.groupby(["family", "node"])["dS"].mean()
    node_averaged_dS_inc = node_averaged_dS_inc.to_frame(name='node_averaged_dS_outlierincluded')
    return node_averaged_dS_inc

def get_nodeaverged_dS_outlierexcluded(df,cutoff = 5):
    df = df[df['dS']<cutoff]
    node_averaged_dS_exc = df.groupby(["family", "node"])["dS"].mean()
    node_averaged_dS_exc = node_averaged_dS_exc.to_frame(name='node_averaged_dS_outlierexcluded')
    return node_averaged_dS_exc

def can_i_run_software(software):
    """
    Test if external software is executable

    :param software: list or string of executable(s)
    :return: 1 (failure) or 0 (success)
    """
    if type(software) == str:
        software = [software]
    ex = 0
    for s in software:
        # codeml needs input otherwise it prompts the user for input, so a dummy
        # file is created
        if s == 'codeml':
            tmp_file = uniq_id()
            with open(tmp_file, 'w') as o:
                o.write('test')
            command = ['codeml', tmp_file]
        elif s == 'prank':
            command = [s, '--help']
        elif s in ['blastp', 'makeblastdb', 'blast', 'muscle', 'i-adhore']:
            command = [s, '-version']
        else:
            command = [s, '--version']
        try:
            sp = subprocess.run(command, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
            logging.info(sp.stdout.decode("utf-8").strip())
        except FileNotFoundError:
            logging.error('{} executable not found!'.format(s))
            ex = 1

        # remove the dummy file
        if s == 'codeml':
            logging.info('codeml found')
            os.remove(tmp_file)
            subprocess.run(['rm', 'rub', 'rst1', 'rst'])
    return ex


def log_subprocess(program, process):
    """
    Log output from a subprocess call to debug log stream

    :param program: program name
    :param process: completed subprocess object
    """
    logging.debug('{} stdout:\n'.format(program) +
                  process.stdout.decode('utf-8'))
    logging.debug('{} stderr:\n'.format(program) +
                  process.stderr.decode('utf-8'))


def get_gfs_for_species(gene_family_dict, gene_pattern):
    """
    Get non-singleton gene families for a species of interest

    :param gene_family_dict: gene family dictionary
    :param species: species of interest
    :return: dictionairy with gene families
    """
    gene_pattern = re.compile(gene_pattern)
    gf_dict = {}

    for key in gene_family_dict.keys():
        if gene_pattern.search(' '.join(gene_family_dict[key])) is not None:
            gf_dict[key] = gene_family_dict[key]

    return gf_dict


def get_sequences(paralog_dict, sequences):
    """
    Fetch sequences from a fasta file or sequence dict and put them in a two
    level dictionairy {gene_family: {gene: seq, gene: seq, ...}, ...}

    :param paralog_dict:
    :return: two-level dictionairy
    """

    if not type(sequences) == dict:
        sequences = read_fasta(sequences, split_on_pipe=True)

    paralog_sequence_dict = {}

    for family in paralog_dict:
        paralog_sequence_dict[family] = {}
        for gene in paralog_dict[family]:
            if gene not in sequences.keys():
                warnings.warn("Gene {} in gene families but not in protein "
                              "fasta!".format(gene))
            else:
                paralog_sequence_dict[family][gene] = sequences[gene]

    return paralog_sequence_dict


def process_gene_families(gene_family_file, ignore_prefix=False):
    """
    Processes a raw gene family file as e.g. from MCL output into a generic
    dictionary structure. MCL raw file consists of one gene family per line,
    including tab separated gene IDs, (without gene family ID !).

    Example::

        gene1   gene2   gene3   gene4   gene5
        gene6   gene7   gene8
        gene9   gene10  gene11
        gene12

    :param gene_family_file: file in the right raw gene family format
        (see above)
    :param ignore_prefix: ignore prefixes (boolean), if the gene contains a '|'
        character in default mode the part preceding the '|' is trimmed of,
        assuming the second part is the gene ID. If ``ignore prefix = True``
        this behavior is suppressed. So with ``ignore_prefix = False``
        ``ath|AT1G10000`` becomes ``AT1G10000`` after processing, with
        ``ignore_prefix = True`` it remains ``ath|AT1G10000``.
    """
    gene_family_dict = {}
    ID = 1

    with open(gene_family_file, 'r') as f:
        for line in f:
            genes = line.strip().split("\t")
            if ignore_prefix:
                if '|' in genes[0]:
                    genes = [gene.split('|')[1] for gene in genes]
            gene_family_dict["GF_{:06d}".format(ID)] = genes
            ID += 1

    return gene_family_dict


def uniq_id():
    """
    Get a unique ID which is not crazy long

    :return: string
    """
    from time import time
    return str(hex(int(time()*10000000))[2:])


def check_dirs(tmp_dir, output_dir, prompt, preserve):
    """
    Check directories needed

    :param tmp_dir: tmp directory
    :param output_dir: output directory
    :param prompt: prompt for overwrites (boolean)?
    :param preserve: preserve MSA files (boolean)?
    :return: nothing
    """
    # Check the tmp directory
    if tmp_dir:
        if os.path.exists(tmp_dir):
            if prompt:
                overwrite = input("tmp directory {} already exists. Overwrite? "
                                  "[y/n]: ".format(tmp_dir))
                if overwrite == 'y':
                    os.system('rm -r {}'.format(os.path.join(tmp_dir)))
                else:
                    print('EXIT')
                    return
            else:
                os.system('rm -r {}'.format(os.path.join(tmp_dir)))
        os.mkdir(tmp_dir)

    # Check the output directory
    if output_dir:
        if os.path.exists(output_dir):
            if prompt:
                overwrite = input("Output directory {} already exists. "
                                  "Overwrite? [y/n]: ".format(output_dir))
                if overwrite == 'y':
                    os.system('rm -r {}'.format(os.path.join(output_dir)))
                else:
                    print('EXIT')
                    return
            else:
                os.system('rm -r {}'.format(os.path.join(tmp_dir)))
        os.mkdir(output_dir)

    # If data should be preserved, make/clean directories
    if preserve:
        if 'msa' not in os.listdir(output_dir):
            os.mkdir(os.path.join(output_dir, 'msa'))

        if 'codeml' not in os.listdir(output_dir):
            os.mkdir(os.path.join(output_dir, 'codeml'))


def read_fasta(fasta_file, prefix=None, split_on_pipe=False,
               split_on_whitespace=True, raw=False):
    """
    Generic fasta file reader. Returns a dictionairy ``{ID: sequence, ID:
    sequence, ...}``.

    :param fasta_file: fasta file
    :param prefix: prefix to add to gene IDs
    :param split_on_pipe: boolean, split gene IDs on '|' character
    :param split_on_whitespace: boolean, split gene IDs on whitespace
    :param raw: boolean, return raw fasta file (why would you want this?)
    :return: sequence dictionary
    """
    invalid_chars = [":", ";", "?"]
    sequence_dict = {}

    with open(fasta_file, 'r') as f:
        if raw:
            return f.read()

        genes = f.read().split(">")
        for gene in genes:
            ID = gene.split("\n")[0]
            for char in invalid_chars:
                if char in ID:
                    raise ValueError(
                        "Gene ID '{0}' contains illegal character '{1}'".format(ID, char))
            if ID != '':
                if split_on_pipe:
                    ID = ID.split("|")[0].strip()
                if split_on_whitespace:
                    ID = ID.split()[0]
                if prefix and prefix != '':
                    ID = prefix + '|' + ID
                sequence = "".join(gene.split("\n")[1:])
                sequence = sequence.replace('*', '')
                sequence_dict[ID] = sequence.upper()

    if '' in sequence_dict.keys():
        del sequence_dict['']
    return sequence_dict


def get_paralogs_fasta(input_fasta, selected_paralogs, output_fasta,
                       pairs=False):
    """
    Get the fasta file associated with the paralogs in a slice from a Ks
    distribution data frame

    :param input_fasta: fasta file
    :param selected_paralogs: Data frame (slice)
    :param output_fasta: output fasta file
    :return: nada
    """
    seqs = read_fasta(input_fasta)

    if not pairs:
        genes = set(selected_paralogs['Paralog1']) | \
                set(selected_paralogs['Paralog2'])
        with open(output_fasta, 'w') as f:
            for gene in list(genes):
                ks_values = list(
                        selected_paralogs[
                            selected_paralogs['Paralog1'] == gene]['Ks']
                )
                ks_values += list(
                        selected_paralogs[
                            selected_paralogs['Paralog2'] == gene]['Ks']
                )
                ks_value, var = mean(ks_values), std(ks_values)
                if gene in seqs.keys():
                    f.write('>{0} mean(Ks)={1:.5f};std(Ks)={2:.5f}\n{3}\n'
                            ''.format(
                            gene, float(ks_value), float(var), seqs[gene]))
                else:
                    logging.warning(
                            'Gene {} not found in fasta file!'.format(gene))

    if pairs:
        for row in selected_paralogs.index:
            p1, p2 = selected_paralogs.loc[row]['Paralog1'], \
                     selected_paralogs.loc[row]['Paralog2']
            if p1 in seqs.keys() and p2 in seqs.keys():
                with open('{0}_{1}_{2:.3f}_{3}'.format(
                        p1, p2, selected_paralogs.loc[row]['Ks'], output_fasta),
                        'w') as o:
                    o.write('>{0}\n{1}\n>{2}\n{3}'.format(
                            p1, seqs[p1], p2, seqs[p2]))
            else:
                logging.warning('Gene not found in fasta file!')

    return


def translate_cds(sequence_dict, skip_invalid=False):
    """
    Just another CDS to protein translater. Will give warnings when in-frame
    stop codons are found, invalid codons are found, or when the sequence length
    is not a multiple of three. Will translate the full sequence or until an
    unspecified or in-frame codon is encountered.

    :param sequence_dict: dictionary with gene IDs and CDS sequences
    :param skip_invalid: bool, skip invalid CDS? (default translates to first
        stop codon or end)
    :return: dictionary with gene IDs and proteins sequences
    """
    # TODO I should just use the Biopython translator
    aa_dict = {
        'ATA': 'I', 'ATC': 'I', 'ATT': 'I', 'ATG': 'M',
        'ACA': 'T', 'ACC': 'T', 'ACG': 'T', 'ACT': 'T',
        'AAC': 'N', 'AAT': 'N', 'AAA': 'K', 'AAG': 'K',
        'AGC': 'S', 'AGT': 'S', 'AGA': 'R', 'AGG': 'R',
        'CTA': 'L', 'CTC': 'L', 'CTG': 'L', 'CTT': 'L',
        'CCA': 'P', 'CCC': 'P', 'CCG': 'P', 'CCT': 'P',
        'CAC': 'H', 'CAT': 'H', 'CAA': 'Q', 'CAG': 'Q',
        'CGA': 'R', 'CGC': 'R', 'CGG': 'R', 'CGT': 'R',
        'GTA': 'V', 'GTC': 'V', 'GTG': 'V', 'GTT': 'V',
        'GCA': 'A', 'GCC': 'A', 'GCG': 'A', 'GCT': 'A',
        'GAC': 'D', 'GAT': 'D', 'GAA': 'E', 'GAG': 'E',
        'GGA': 'G', 'GGC': 'G', 'GGG': 'G', 'GGT': 'G',
        'TCA': 'S', 'TCC': 'S', 'TCG': 'S', 'TCT': 'S',
        'TTC': 'F', 'TTT': 'F', 'TTA': 'L', 'TTG': 'L',
        'TAC': 'Y', 'TAT': 'Y', 'TAA': '',  'TAG': '',
        'TGC': 'C', 'TGT': 'C', 'TGA': '',  'TGG': 'W',
    }
    protein_dict = {}

    with ProgressBar(
            max_value=len(sequence_dict.keys()) + 1, redirect_stdout=True,
            redirect_stderr=True
    ) as pb:
        j = 0
        total = 0
        for key, val in sequence_dict.items():
            j += 1
            aa_seq = ''
            if len(val) % 3 != 0:
                print('Sequence length != multiple of 3 for {}!'.format(key))
                total += 1
            invalid = False
            for i in range(0, len(val), 3):
                if val[i:i + 3] not in aa_dict.keys():
                    print('Invalid codon {0:>3} in {1}'.format(val[i:i+3], key))
                    invalid = True
                    total += 1
                    break
                else:
                    if aa_dict[val[i:i + 3]] == '' and i+3 != len(val):
                        print('In-frame STOP codon in {0} at position {1}:{2}'
                              ''.format(key, i, i+3))
                        invalid = True
                        total += 1
                        break
                    aa_seq += aa_dict[val[i:i + 3]]
            if invalid and skip_invalid:
                continue
            protein_dict[key] = aa_seq
            pb.update(j)
    logging.warning("There were {} warnings during translation".format(total))
    return protein_dict


def write_fasta(seq_dict, output_file):
    """
    Write a sequence dictionary to a fasta file.

    :param seq_dict: sequence dictionary, see :py:func:`read_fasta`
    :param output_file: output file name
    """
    with open(output_file, 'w') as o:
        for key, val in seq_dict.items():
            o.write('>' + key + '\n')
            o.write(val + '\n')
    return output_file


def filter_one_vs_one_families(gene_families, s1, s2):
    """
    Filter one-vs-one ortholog containing families for two given species.

    :param gene_families: gene families fil in raw MCL format, see
        :py:func:`process_gene_families`
    :param s1: species 1 prefix
    :param s2: species 2 prefix
    :return: one-vs.-one ortholog containing gene families.
    """
    to_delete = []
    for key, val in gene_families.items():
        count = 0
        for gene in val:
            prefix = gene.split('|')[0]
            if prefix == s1 or prefix == s2:
                count += 1
        if count != 2:
            to_delete.append(key)
    for k in to_delete:
        del gene_families[k]
    return gene_families


def _random_color():
    """
    Generate a random hex color
    """

    def r(): return random.randint(0, 255)

    return '#%02X%02X%02X' % (r(), r(), r())


class Genome:
    """
    Class that represents a structural annotation.
    Collects several nice data structures for a genome and parsers for various
    genomic data file formats (e.g. gff, fasta, ...)
    """

    def __init__(self):
        """
        Genome.genome: dictionary with a full representation
        Genome.gene_lists: dictionary with ordered lists of genes per
        chromosome (as for I-ADHoRe)
        """
        self.parent_file = None
        self.genome = {}
        self.gene_lists = {}
        self.colors = {}

    def parse_plaza_gff(self, gff_file, keyword='mRNA', id_string='Parent'):
        """
        Parse a PLAZA annotation file into a genome dictionary

        :param gff_file: input gff (PLAZA style)
        :param keyword: keyword for elements to parse out
        :param id_string: keyword for retrieving the gene ID from the 9th column
        """
        self.parent_file = gff_file

        with open(gff_file, 'r') as f:
            for line in f:
                if line.startswith('#'):
                    continue
                line = line.strip().split('\t')

                if line[2] == keyword:
                    chromosome = line[0]
                    start = line[3]
                    stop = line[4]
                    orientation = line[6]
                    gene_l = line[8].split(';')
                    gene_dict = {x.split('=')[0]: x.split('=')[1] for x in
                                 gene_l if len(x.split('=')) == 2}

                    if chromosome not in self.genome:
                        self.genome[chromosome] = {}
                        self.gene_lists[chromosome] = []
                        self.colors[chromosome] = _random_color()

                    self.genome[chromosome][gene_dict[id_string]] = {
                        'orientation': orientation, 'start': start,
                        'stop': stop}
                    self.gene_lists[chromosome].append(
                            (gene_dict[id_string], orientation, start, stop))
        return

    def karyotype_json(self, out_file='genome.json'):
        """
        Generate karyotype data file in json format (as per Circos.js/d3.js)

        :param out_file: output file name
        """
        karyotype = []
        for chrom in self.gene_lists.keys():
            # approximate chromosome length
            coordinates = [int(x[2]) for x in self.gene_lists[chrom]]
            coordinates += [int(x[3]) for x in self.gene_lists[chrom]]
            length = max(coordinates) - min(coordinates)
            karyotype.append({'id': chrom, 'label': chrom,
                              'color': self.colors[chrom], 'len': length})

        if out_file:
            with open(out_file, 'w') as f:
                json.dump(karyotype, f)

        else:
            return json.dumps(karyotype)


class gaussian_kde(object):
    """
    from: https://stackoverflow.com/questions/27623919/weighted-gaussian-kernel-
    density-estimation-in-python

    Representation of a kernel-density estimate using Gaussian kernels.

    Kernel density estimation is a way to estimate the probability density
    function (PDF) of a random variable in a non-parametric way.
    `gaussian_kde` works for both uni-variate and multi-variate data.   It
    includes automatic bandwidth determination.  The estimation works best for
    a unimodal distribution; bimodal or multi-modal distributions tend to be
    oversmoothed.

    Parameters
    ----------
    dataset : array_like
        Datapoints to estimate from. In case of univariate data this is a 1-D
        array, otherwise a 2-D array with shape (# of dims, # of data).
    bw_method : str, scalar or callable, optional
        The method used to calculate the estimator bandwidth.  This can be
        'scott', 'silverman', a scalar constant or a callable.  If a scalar,
        this will be used directly as `kde.factor`.  If a callable, it should
        take a `gaussian_kde` instance as only parameter and return a scalar.
        If None (default), 'scott' is used.  See Notes for more details.
    weights : array_like, shape (n, ), optional, default: None
        An array of weights, of the same shape as `x`.  Each value in `x`
        only contributes its associated weight towards the bin count
        (instead of 1).

    Attributes
    ----------
    dataset : ndarray
        The dataset with which `gaussian_kde` was initialized.
    d : int
        Number of dimensions.
    n : int
        Number of datapoints.
    neff : float
        Effective sample size using Kish's approximation.
    factor : float
        The bandwidth factor, obtained from `kde.covariance_factor`, with which
        the covariance matrix is multiplied.
    covariance : ndarray
        The covariance matrix of `dataset`, scaled by the calculated bandwidth
        (`kde.factor`).
    inv_cov : ndarray
        The inverse of `covariance`.

    Methods
    -------
    kde.evaluate(points) : ndarray
        Evaluate the estimated pdf on a provided set of points.
    kde(points) : ndarray
        Same as kde.evaluate(points)
    kde.pdf(points) : ndarray
        Alias for ``kde.evaluate(points)``.
    kde.set_bandwidth(bw_method='scott') : None
        Computes the bandwidth, i.e. the coefficient that multiplies the data
        covariance matrix to obtain the kernel covariance matrix.
        .. versionadded:: 0.11.0
    kde.covariance_factor : float
        Computes the coefficient (`kde.factor`) that multiplies the data
        covariance matrix to obtain the kernel covariance matrix.
        The default is `scotts_factor`.  A subclass can overwrite this method
        to provide a different method, or set it through a call to
        `kde.set_bandwidth`.

    Notes
    -----
    Bandwidth selection strongly influences the estimate obtained from the KDE
    (much more so than the actual shape of the kernel).  Bandwidth selection
    can be done by a "rule of thumb", by cross-validation, by "plug-in
    methods" or by other means; see [3]_, [4]_ for reviews.  `gaussian_kde`
    uses a rule of thumb, the default is Scott's Rule.

    Scott's Rule [1]_, implemented as `scotts_factor`, is::

        n**(-1./(d+4)),

    with ``n`` the number of data points and ``d`` the number of dimensions.
    Silverman's Rule [2]_, implemented as `silverman_factor`, is::

        (n * (d + 2) / 4.)**(-1. / (d + 4)).

    Good general descriptions of kernel density estimation can be found in [1]_
    and [2]_, the mathematics for this multi-dimensional implementation can be
    found in [1]_.

    References
    ----------
    .. [1] D.W. Scott, "Multivariate Density Estimation: Theory, Practice, and
           Visualization", John Wiley & Sons, New York, Chicester, 1992.
    .. [2] B.W. Silverman, "Density Estimation for Statistics and Data
           Analysis", Vol. 26, Monographs on Statistics and Applied Probability,
           Chapman and Hall, London, 1986.
    .. [3] B.A. Turlach, "Bandwidth Selection in Kernel Density Estimation: A
           Review", CORE and Institut de Statistique, Vol. 19, pp. 1-33, 1993.
    .. [4] D.M. Bashtannyk and R.J. Hyndman, "Bandwidth selection for kernel
           conditional density estimation", Computational Statistics & Data
           Analysis, Vol. 36, pp. 279-298, 2001.

    Examples
    --------
    Generate some random two-dimensional data:

    >>> from scipy import stats
    >>> def measure(n):
    >>>     "Measurement model, return two coupled measurements."
    >>>     m1 = np.random.normal(size=n)
    >>>     m2 = np.random.normal(scale=0.5, size=n)
    >>>     return m1+m2, m1-m2

    >>> m1, m2 = measure(2000)
    >>> xmin = m1.min()
    >>> xmax = m1.max()
    >>> ymin = m2.min()
    >>> ymax = m2.max()

    Perform a kernel density estimate on the data:

    >>> X, Y = np.mgrid[xmin:xmax:100j, ymin:ymax:100j]
    >>> positions = np.vstack([X.ravel(), Y.ravel()])
    >>> values = np.vstack([m1, m2])
    >>> kernel = stats.gaussian_kde(values)
    >>> Z = np.reshape(kernel(positions).T, X.shape)

    Plot the results:

    >>> import matplotlib.pyplot as plt
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(111)
    >>> ax.imshow(np.rot90(Z), cmap=plt.cm.gist_earth_r,
    ...           extent=[xmin, xmax, ymin, ymax])
    >>> ax.plot(m1, m2, 'k.', markersize=2)
    >>> ax.set_xlim([xmin, xmax])
    >>> ax.set_ylim([ymin, ymax])
    >>> plt.show()

    """

    def __init__(self, dataset, bw_method=None, weights=None):
        self.dataset = np.atleast_2d(dataset)
        if not self.dataset.size > 1:
            raise ValueError("`dataset` input should have multiple elements.")
        self.d, self.n = self.dataset.shape

        if weights is not None:
            self.weights = weights / np.sum(weights)
        else:
            self.weights = np.ones(self.n) / self.n

        # Compute the effective sample size
        # http://surveyanalysis.org/wiki/Design_Effects_and_Effective_Sample_
        # Size#Kish.27s_approximate_formula_for_computing_effective_sample_size
        self.neff = 1.0 / np.sum(self.weights ** 2)

        self.set_bandwidth(bw_method=bw_method)

    def evaluate(self, points):
        """Evaluate the estimated pdf on a set of points.

        Parameters
        ----------
        points : (# of dimensions, # of points)-array
            Alternatively, a (# of dimensions,) vector can be passed in and
            treated as a single point.

        Returns
        -------
        values : (# of points,)-array
            The values at each point.

        Raises
        ------
        ValueError : if the dimensionality of the input points is different than
                     the dimensionality of the KDE.

        """
        points = np.atleast_2d(points)

        d, m = points.shape
        if d != self.d:
            if d == 1 and m == self.d:
                # points was passed in as a row vector
                points = np.reshape(points, (self.d, 1))
                m = 1
            else:
                msg = "points have dimension %s, dataset has dimension %s" % \
                      (d, self.d)
                raise ValueError(msg)

        # compute the normalised residuals
        chi2 = cdist(points.T, self.dataset.T, 'mahalanobis',
                     VI=self.inv_cov) ** 2
        # compute the pdf
        result = np.sum(np.exp(-.5 * chi2) * self.weights,
                        axis=1) / self._norm_factor

        return result

    __call__ = evaluate

    def scotts_factor(self):
        return np.power(self.neff, -1. / (self.d + 4))

    def silverman_factor(self):
        return np.power(self.neff * (self.d + 2.0) / 4.0, -1. / (self.d + 4))

    #  Default method to calculate bandwidth, can be overwritten by subclass
    covariance_factor = scotts_factor

    def set_bandwidth(self, bw_method=None):
        """Compute the estimator bandwidth with given method.

        The new bandwidth calculated after a call to `set_bandwidth` is used
        for subsequent evaluations of the estimated density.

        Parameters
        ----------
        bw_method : str, scalar or callable, optional
            The method used to calculate the estimator bandwidth.  This can be
            'scott', 'silverman', a scalar constant or a callable.  If a
            scalar, this will be used directly as `kde.factor`.  If a callable,
            it should take a `gaussian_kde` instance as only parameter and
            return a scalar.  If None (default), nothing happens; the current
            `kde.covariance_factor` method is kept.

        Notes
        -----
        .. versionadded:: 0.11

        Examples
        --------
        >>> x1 = np.array([-7, -5, 1, 4, 5.])
        >>> kde = stats.gaussian_kde(x1)
        >>> xs = np.linspace(-10, 10, num=50)
        >>> y1 = kde(xs)
        >>> kde.set_bandwidth(bw_method='silverman')
        >>> y2 = kde(xs)
        >>> kde.set_bandwidth(bw_method=kde.factor / 3.)
        >>> y3 = kde(xs)

        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(111)
        >>> ax.plot(x1, np.ones(x1.shape) / (4. * x1.size), 'bo',
        ...         label='Data points (rescaled)')
        >>> ax.plot(xs, y1, label='Scott (default)')
        >>> ax.plot(xs, y2, label='Silverman')
        >>> ax.plot(xs, y3, label='Const (1/3 * Silverman)')
        >>> ax.legend()
        >>> plt.show()

        """
        if bw_method is None:
            pass
        elif bw_method == 'scott':
            self.covariance_factor = self.scotts_factor
        elif bw_method == 'silverman':
            self.covariance_factor = self.silverman_factor
        elif np.isscalar(bw_method) and not isinstance(bw_method, str):
            self._bw_method = 'use constant'
            self.covariance_factor = lambda: bw_method
        elif callable(bw_method):
            self._bw_method = bw_method
            self.covariance_factor = lambda: self._bw_method(self)
        else:
            msg = "`bw_method` should be 'scott', 'silverman', a scalar " \
                  "or a callable."
            raise ValueError(msg)

        self._compute_covariance()

    def _compute_covariance(self):
        """Computes the covariance matrix for each Gaussian kernel using
        covariance_factor().
        """
        self.factor = self.covariance_factor()
        # Cache covariance and inverse covariance of the data
        if not hasattr(self, '_data_inv_cov'):
            # Compute the mean and residuals
            _mean = np.sum(self.weights * self.dataset, axis=1)
            _residual = (self.dataset - _mean[:, None])
            # Compute the biased covariance
            self._data_covariance = np.atleast_2d(
                    np.dot(_residual * self.weights, _residual.T))
            # Correct for bias (http://en.wikipedia.org/wiki/Weighted_arithmetic
            # _mean#Weighted_sample_covariance)
            self._data_covariance /= (1 - np.sum(self.weights ** 2))
            self._data_inv_cov = np.linalg.inv(self._data_covariance)

        self.covariance = self._data_covariance * self.factor ** 2
        self.inv_cov = self._data_inv_cov / self.factor ** 2
        self._norm_factor = np.sqrt(
                np.linalg.det(2 * np.pi * self.covariance))  # * self.n



# TODO: implement pair-by-pair codeml analysis for a given input alignment.
import pandas as pd
import numpy as np
import subprocess as sp
import logging
import os
import re
from io import StringIO
from Bio import Phylo
from Bio.Align import MultipleSeqAlignment
import seaborn as sns
import matplotlib.pyplot as plt


GAPS = ["-", "?", "X"]  # these should be treated as gaps

def _strip_gaps(aln):
    new_aln = aln[:,0:0]
    for j in range(aln.get_alignment_length()):
        if any([x in GAPS for x in aln[:,j]]):
            continue
        else:
            new_aln += aln[:,j:j+1]
    return new_aln

def _all_pairs(aln):
    pairs = []
    for i in range(len(aln)-1):
        for j in range(i+1, len(aln)):
            pairs.append([aln[i].id, aln[j].id])
    return pairs

def _write_aln_codeml(aln, fname):
    with open(fname, "w") as f:
        f.write("{} {}\n".format(len(aln), aln.get_alignment_length()))
        for s in aln:
            f.write("{}\n".format(s.id))
            f.write("{}\n".format(s.seq))

def _parse_pairwise(codeml_out):
    with open(codeml_out, "r") as f:
        content = f.read()
    x = content.split("pairwise comparison")[2].split("\n")
    y = [x for x in map(lambda i: _parse_pair(x, i), range(3, len(x), 7))]
    d = pd.DataFrame.from_dict(y).set_index("pair")
    return d

def _parse_pair(lines, start):
    x = lines[start].split()
    a = x[1].strip("()")
    b = x[4].strip("()")
    l = float(lines[start+1].split("=")[1].strip())
    x = lines[start+4].split("=")
    y = [x[0]]
    for elem in x[1:]:
        y += elem.split()
    x = {y[i].strip(): float(y[i+1].strip()) for i in range(0, len(y)-1, 2)}
    p = "__".join(sorted([a, b]))
    x.update({"pair": p, "gene1": a, "gene2": b, "l": l})
    return x

def _parse_tree(codeml_out,kstree_dir, katree_dir, wtree_dir, prefix, parentdir):
    content = []
    with open(codeml_out, "r") as f:
        for i in f.readlines(): content.append(i)
    ds,dn,w = content[-9],content[-7],content[-4].replace(' #',': ')
    os.chdir(parentdir)
    _write_basic(ds,os.path.join(kstree_dir,'{}.dS.tree'.format(prefix)))
    _write_basic(dn,os.path.join(katree_dir,'{}.dN.tree'.format(prefix)))
    _write_basic(w,os.path.join(wtree_dir,'{}.w.tree'.format(prefix)))
    #handle_ds,handle_dn,handle_w = StringIO(ds),StringIO(dn),StringIO(w)
    #tree_ds,tree_dn,tree_w = Phylo.read(handle_ds, "newick"),Phylo.read(handle_dn, "newick"),Phylo.read(handle_w, "newick")
    tree_ds,tree_dn,tree_w = Phylo.read(os.path.join(kstree_dir,'{}.dS.tree'.format(prefix)), "newick"),Phylo.read(os.path.join(katree_dir,'{}.dN.tree'.format(prefix)), "newick"),Phylo.read(os.path.join(wtree_dir,'{}.w.tree'.format(prefix)), "newick")
    drawtree(tree_ds,os.path.join(kstree_dir,prefix+".dS.tree.pdf"))
    drawtree(tree_dn,os.path.join(katree_dir,prefix+".dN.tree.pdf"))
    drawtree(tree_w,os.path.join(wtree_dir,prefix+".w.tree.pdf"))
    return ds,dn,w

def drawtree(tree,fname):
    fig,ax = plt.subplots()
    Phylo.draw(tree,do_show=False,axes=ax)
    sns.despine(offset=1)
    fig.savefig(fname)
    plt.close()

def _write_basic(content,fname):
    with open(fname,'w') as f: f.write(content)

def check_noneresult(codeml_out):
    with open(codeml_out, "r") as f: content = f.read()
    if "pairwise comparison" not in content:
        return False
    else:
        return True

def _run_codeml_tree(exe, control_file, out_file, prefix, kstree_dir, katree_dir, wtree_dir, parentdir, preserve=False, times=1):
    #logging.debug("Performing codeml {} times".format(times))
    #max_results = None
    #max_likelihood = None
    #Noresults = False
    for i in range(times):
        #logging.debug("Codeml iteration {0} for {1}".format(str(i+1), control_file))
        sp.run([exe, control_file], stdout=sp.PIPE)
        #sp.run(['rm', '2ML.dN', '2ML.dS', '2ML.t', '2NG.dN', '2NG.dS','2NG.t', 'rst', 'rst1', 'rub'], stdout=sp.PIPE, stderr=sp.PIPE)
        if not os.path.isfile(out_file):
            raise FileNotFoundError('Codeml output file not found')
    ds,dn,w = _parse_tree(out_file,kstree_dir,katree_dir,wtree_dir,prefix,parentdir)
    #_write_basic(ds,os.path.join(kstree_dir,'{}.dS.tree'.format(prefix)))
    #_write_basic(dn,os.path.join(katree_dir,'{}.dN.tree'.format(prefix)))
    #_write_basic(w,os.path.join(wtree_dir,'{}.w.tree'.format(prefix)))


def _run_codeml(exe, control_file, out_file, preserve=False, times=1):
    """
    Run codeml assuming all necessary files are written and we are in the
    directory with those files. Set `preserve` to true to keep intermediary
    files. When `times > 1`, codeml will be run `times` times and the best
    result (highest likelihood) will be returned.
    """
    logging.debug("Performing codeml {} times".format(times))
    max_results = None 
    max_likelihood = None
    Noresults = False
    for i in range(times):
        logging.debug("Codeml iteration {0} for {1}".format(str(i+1), control_file))
        sp.run([exe, control_file], stdout=sp.PIPE)
        sp.run(['rm', '2ML.dN', '2ML.dS', '2ML.t', '2NG.dN', '2NG.dS',
            '2NG.t', 'rst', 'rst1', 'rub'], stdout=sp.PIPE, stderr=sp.PIPE)
        if not os.path.isfile(out_file):
            raise FileNotFoundError('Codeml output file not found')
        if not check_noneresult(out_file):
            Noresults = True
            break
        results = _parse_pairwise(out_file)
        likelihood = sum(results["l"])
        if not max_likelihood or likelihood > max_likelihood:
            max_likelihood = likelihood
            max_results = results
    if Noresults:
        if not preserve:
            os.remove(out_file)
        return None
    logging.debug('Best MLE: ln(L) = {}'.format(max_likelihood))
    #os.remove(control_file)
    if not preserve:
        os.remove(out_file)
    return max_results


class Codeml:
    """
    Class for codeml (PAML Yang 2007) python wrapper. Defines the controle file
    and enables running codeml from within python in one line of code.

    :param codeml: path to codeml executable (by default will look for codeml
        in the system PATH)
    :param tmp: path to temporary directory, will default to the current working
        directory ('./')
    :param id: filename prefix for output/tmp files
    :param kwargs: other codeml keyword arguments (see PAML user guide)::

        'seqfile': None,
        'outfile': self.out_file,
        'noisy': 0,
        'verbose': 0,
        'runmode': -2,
        'seqtype': 1,
        'CodonFreq': 2,
        'clock': 0,
        'aaDist': 0,
        'aaRatefile': 'dat/jones.dat',
        'model': 0,
        'NSsites': 0,
        'icode': 0,
        'Mgene': 0,
        'fix_kappa': 0,
        'kappa': 2,
        'fix_omega': 0,
        'omega': .4,
        'fix_alpha': 1,
        'alpha': 0,
        'Malpha': 0,
        'ncatG': 8,
        'getSE': 0,
        'RateAncestor': 1,
        'Small_Diff': .5e-6,
        'cleandata': 1,
        'method': 0
    """
    def __init__(self, aln, exe='codeml', tmp='./', prefix='codeml', treefile=None, **kwargs):
        """
        Codeml wrapper init. Initializes the default control file for Ks
        analysis as proposed by Vanneste et al. (2013). Takes as keyword
        arguments the options from the normal codeml distribution. Control
        settings are stored in a dictionary that can be accessed with the
        `.control` attribute

        :param aln: alignment file
        :param exe: path to codeml executable (by default will look for
            codeml in the system PATH)
        :param tmp: path to temporary directory, will default to the current
            working directory ('./')
        :param prefix: filename prefix for output/tmp files
        :param kwargs: any codeml control option (see PAML user guide)
        """
        if not os.path.isdir(tmp):
            raise NotADirectoryError('tmp directory {} not found!'.format(tmp))
        self.prefix = prefix
        self.aln = aln
        self.exe = exe
        self.tmp = tmp
        self.treefile = treefile
        self.control_file = self.prefix + '.ctrl'
        self.aln_file = self.prefix + '.cdsaln'
        self.out_file = self.prefix + '.codeml'
        self.control = {
            'seqfile': self.aln_file, 
            'outfile': self.out_file,
            'noisy': 0, 
            'verbose': 0, 
            'runmode': -2, 
            'seqtype': 1,
            'CodonFreq': 2, 
            'clock': 0, 
            'aaDist': 0,
            'aaRatefile': 'dat/jones.dat', 
            'model': 0, 
            'NSsites': 0, 
            'icode': 0,
            'Mgene': 0, 
            'fix_kappa': 0,
            'kappa': 2, 
            'fix_omega': 0, 
            'omega': .4, 
            'fix_alpha': 1, 
            'alpha': 0,
            'Malpha': 0, 
            'ncatG': 8,
            'getSE': 0, 
            'RateAncestor': 1, 
            'Small_Diff': .5e-6, 
            'cleandata': 1,
            'method': 0}
        # update the control with kwargs
        for x in kwargs.keys():
            if x not in self.control:
                raise KeyError("{} is not a valid codeml param.".format(x))
            else:
                self.control[x] = kwargs[x]

    def __str__(self):
        """
        String method for Codeml wrapper, prints current control settings

        :return: string representation of the control file
        """
        x = ['{0} = {1}\n'.format(k, v) for (k,v) in sorted(self.control.items())]
        return "\n".join(x)

    def write_ctrl(self):
        with open(self.control_file, "w") as f:
            f.write(str(self))

    # We should output the pairs for which we couldn't estimate Ks separately,
    # so that we can make a nan_result for those, but work with those we could
    # estimate without much trouble
    def run_codeml(self, kstree_dir=None, katree_dir=None, wtree_dir=None, **kwargs):
        """
        Run codeml on the full alignment. This will exclude all gap-containing
        columns, which may lead to a significant loss of data.
        """
        stripped_aln = _strip_gaps(self.aln)  # codeml does this anyway
        if stripped_aln.get_alignment_length() == 0:
            logging.warning("Stripped alignment length == 0 for {}".format(self.prefix))
            return None, _all_pairs(self.aln)
        parentdir = os.path.abspath(os.curdir)  # where we are currently
        os.chdir(self.tmp)  # go to tmpdir
        if not (self.treefile is None):
            self.control['treefile'] = self.treefile
            self.control['model'] = 1
            self.control['runmode'] = 0
            self.control['fix_blength'] = 1
        self.write_ctrl() 
        _write_aln_codeml(self.aln, self.aln_file)
        if not (self.treefile is None):
            _run_codeml_tree(self.exe, self.control_file, self.out_file, self.prefix, kstree_dir, katree_dir, wtree_dir,parentdir,**kwargs)
            return
        results = _run_codeml(self.exe, self.control_file, self.out_file, **kwargs)
        os.chdir(parentdir)
        if results is None:
            logging.warning("No codeml result for {} due to no resolved nucleotides".format(self.prefix))
            return None, _all_pairs(self.aln)
        else:
            return results, []

    def run_codeml_pairwise(self, **kwargs):
        """
        Run codeml on each pair of the alignment separately.
        """
        parentdir = os.path.abspath(os.curdir)  # where we are currently
        os.chdir(self.tmp)  # go to tmpdir
        results = []
        no_results = []
        for i in range(len(self.aln)-1):
            for j in range(i+1, len(self.aln)):
                pair = MultipleSeqAlignment([self.aln[i], self.aln[j]])
                stripped_pair = _strip_gaps(pair)
                if stripped_pair.get_alignment_length() == 0:
                    no_results.append([p.id for p in pair])
                else:
                    self.write_ctrl() 
                    _write_aln_codeml(stripped_pair, self.aln_file)
                    tmp_result = _run_codeml(self.exe, self.control_file,self.out_file, **kwargs)
                    if tmp_result is None:
                        no_results.append([p.id for p in pair])
                    else:
                        results.append(tmp_result)
        if len(no_results) > 0:
            logging.warning("Alignment length of 0 for {} pairs in {}".format(
                len(no_results), self.prefix))
        os.chdir(parentdir)
        return pd.concat(results), no_results




import pandas as pd
import fastcluster
import numpy as np
from Bio import Phylo

def cluster_ks(df):
    """
    Cluster a data frame with Ks estimates using average-linkage hierarchical
    clustering.
    """
    X = pairwise_matrix(df, "dS")
    Y = alc_cluster(X)
    T = to_phylo(Y, X)
    return T

def alc_cluster(df, nanpolicy=1000):
    # fill NaN values with something larger than all the rest, not a
    # foolproof approach, but should be reasonable in most cases
    if np.any(np.isnan(df)):
        logging.warning("Data contains NaN values, replaced by "+str(nanpolicy))
        df.fillna(nanpolicy, inplace=True)
    return fastcluster.average(df)

def pairwise_matrix(df, col):
    d = {}
    for p in df.index:
        g1 = df.loc[p].gene1
        g2 = df.loc[p].gene2
        x  = df.loc[p][col]
        if not g1 in d:
            d[g1] = {}
        if not g2 in d:
            d[g2] = {}
        d[g1][g2] = x
        d[g2][g1] = x
    df = pd.DataFrame.from_dict(d).fillna(0.)
    # df[df.index] ensures the result is a proper (symmetric) distance matrix
    return df[df.index]

def to_phylo(clustering, df):
    n = len(df.index)
    heights = {i: 0 for i in range(n)}
    clades = [Phylo.Newick.Clade(0, df.index[i]) for i in range(n)]
    for row in clustering:
        i = int(row[0])
        j = int(row[1])
        c1 = clades[i]
        c2 = clades[j]
        height = row[2] 
        newnode = Phylo.Newick.Clade(0, n)
        c1.branch_length = height - heights[i]
        c2.branch_length = height - heights[j]
        newnode.clades.extend([c1, c2])
        heights[n] = height
        clades.append(newnode)
        n += 1
    return clades[-1]
    



import pandas as pd
from Bio import SeqIO
from Bio.Data.CodonTable import TranslationError
import logging
from collections import Counter
import os
from joblib import Parallel, delayed
from tqdm import tqdm,trange
import subprocess as sp
from Bio import AlignIO
from Bio.Alphabet import IUPAC
from Bio import Phylo
from ete3 import PhyloTree

def _mkdir(dirname):
    if not os.path.isdir(dirname) :
        os.mkdir(dirname)
    return dirname

def fetcher(config):
    """
    Return the essential info in dict
    """
    para_dict = {}
    getinfo = lambda x : (x[0],x[1].strip())
    with open(config,'r') as f:
        for line in f.readlines():
            key,value = getinfo(line.split('\t')[0:2])
            para_dict[key] = value
    return para_dict

def reporter(dic):
    """
    Log out the info
    """
    for key,value in dic.items(): logging.info("{0}\t{1}".format(key,value))

def listdir(data,align=False):
    """
    Return the path to sequence files in a list
    """
    parent = os.getcwd()
    os.chdir(data)
    y = lambda x:os.path.join(data,x)
    files_clean = [y(i) for i in os.listdir() if i!="__pycache__"]
    if not align:
        logging.info("In total {} sequence files are found".format(len(files_clean)))
        logging.info(", ".join([os.path.basename(i) for i in files_clean]))
    os.chdir(parent)
    return files_clean

def write_seq(fam,og,seq,seqtype):
    """
    og is a series
    """
    og = og.dropna()
    with open(fam+'.'+seqtype,'w') as f:
        for sp,gids in og.items():
            for gid in gids.split(', '): f.write('>{0}\n{1}\n'.format(gid,seq[sp][gid].seq))

def write_seq_translate(fam,og,seq,to_stop,cds):
    """
    og is a series
    """
    og = og.dropna()
    with open(fam+'.'+'pep','w') as f:
        for sp,gids in og.items():
            for gid in gids.split(', '): f.write('>{0}\n{1}\n'.format(gid,seq[sp][gid].translate(to_stop=to_stop,cds=cds,id=gid).seq))

def write_seq_translate_fammode(fam,fseq,to_stop,cds,pepdir):
    """
    original seq translation
    """
    with open(os.path.join(pepdir,fam+'.'+'pep'),'w') as f:
        for record in SeqIO.parse(fseq, 'fasta'):
            f.write(">{0}\n{1}\n".format(record.id,record.translate(to_stop=to_stop,cds=cds,id=record.id).seq))

def find_singleton(og):
    """
    return True or False
    """
    og = og.dropna()
    num = 0
    for seqs in og.values:
        num += len(seqs.split(', '))
        if num > 1:
            return False
    return True

def deal_options(options):
    Options = []
    for i in options:
        if len(i.split(' '))==1: Options.append(i)
        else: Options+=i.split(' ')
    return Options

def mafft(fseq,options,faln):
    if options == '': cmd = ["mafft"] + [fseq]
    else: cmd = ["mafft"] + options + [fseq]
    out = sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    with open(faln, 'w') as f: f.write(out.stdout.decode('utf-8'))

def muscle(fseq,options,faln):
    if options == '': cmd = ["muscle"] + ['-in',fseq] + ['-out',faln]
    else: cmd = ["muscle"] + ['-in',fseq] + ['-out',faln] + options
    sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)

def prank(fseq,options,faln):
    if options == '': cmd = ["prank"] + ['-d='+fseq] + ['-o='+faln]
    else: cmd = ["prank"] + ['-d='+fseq] + ['-o='+faln] + options
    sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)

def iqtree(faln,options):
    cmd = ["iqtree", "-s", faln] + options
    sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    if os.path.exists(faln+".treefile"): return faln+".treefile"

def iqtree2(faln,options):
    cmd = ["iqtree2", "-s", faln] + options
    sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    if os.path.exists(faln+".treefile"): return faln+".treefile"

def fasttree(faln,options):
    cmd = ["FastTree"] + options + ["-out", faln+'.FastTree', faln]
    sp.run(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    if os.path.exists(faln+".FastTree"): return faln+'.FastTree'

def mrbayes(faln,options):
    faln_nex = faln+".nexus"
    AlignIO.convert(faln, 'fasta', faln_nex, 'nexus', IUPAC.extended_protein)
    conf =  faln_nex+".config.mb"
    logf = faln_nex+".mb.log"
    bashf = faln_nex+".bash.mb"
    with open(conf,"w") as f:
        if 'set' in options: f.write('set'+' '+' '.join(options['set'])+'\n')
        else: f.write("set autoclose=yes nowarn=yes\n")
        f.write("execute {}\n".format(os.path.basename(faln_nex)))
        if 'prset' in options: f.write('prset'+' '+' '.join(options['prset'])+'\n')
        else: f.write("prset ratepr=variable\n")
        if 'lset' in options: f.write('lset'+' '+' '.join(options['lset'])+'\n')
        else: f.write("lset rates=gamma\n")
        if 'mcmcp' in options: f.write('mcmcp'+' '+' '.join(options['mcmcp'])+'\n')
        else: f.write("mcmcp diagnfreq=100 samplefreq=10\n")
        if 'mcmc' in options: f.write('mcmc'+' '+' '.join(options['mcmc'])+'\n')
        else: f.write("mcmc ngen=1100 savebrlens=yes nchains=1\n")
        f.write("sumt\nsump\nquit\n")
    with open(bashf,"w") as f:
        f.write('mb <{0}> {1}'.format(conf,logf))
    mb_cmd = ["sh", bashf]
    sp.run(mb_cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    with open('rm.sh',"w") as f: f.write("rm *.bash.mb")
    rm_cmd = ['sh', 'rm.sh']
    sp.run(rm_cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    rm_cmd = ['rm', 'rm.sh']
    sp.run(rm_cmd, stdout=sp.PIPE, stderr=sp.PIPE)
    if os.path.exists(faln+".nexus.con.tre"): return os.path.abspath(faln+".nexus.con.tre")

def labelinternalnodes(tree):
    Tree = Phylo.read(tree,'newick')
    clade_map = {}
    for indice,clade in enumerate(Tree.get_nonterminals()):
        clade.confidence = None
        clade.name = "N" + str(indice)
        clade_map[clade.name] = []
        for tip in clade.get_terminals(): clade_map[clade.name].append(tip.name)
    Phylo.write(Tree,tree+'.labelled','newick')
    with open(tree+".nodemap","w") as f:
        for key,value in clade_map.items(): f.write(key+"\t"+", ".join(value)+"\n")
        for clade in Tree.get_terminals(): f.write(clade.name+"\t"+clade.name+"\n")
    return tree+".nodemap"

def intoffloat(Ixx):
    if all([i.confidence==1 for i in Ixx if i.confidence is not None]):
        return "FT"
    else:
        Float = False
        for i in Ixx:
            if type(i.confidence) is float:
                Float = True
                break
        if Float: return "FT"
        else: return "IQ"

def getbssubtree(tree,threshold):
    Ixx = tree.get_nonterminals()
    treetp = intoffloat(Ixx)
    Threshold = threshold if treetp == "FT" else threshold*100
    for i in Ixx:
        if i.confidence is not None:
            if i.confidence < Threshold:
                for j in i.get_terminals():
                    if tree.get_path(j, terminal=True):
                        tree.prune(j)
    return tree

def prunepolytomy(genetree):
    for i in genetree.get_nonterminals():
        if len(i.clades) > 2:
            for j in i.get_terminals():
                if genetree.get_path(j, terminal=True): genetree.prune(j)
    return genetree

def mb2nw(tree):
    mb_out_content,linenumber = [],0
    with open(tree,"r") as f:
        lines = f.readlines()
        for line in lines:
            linenumber += 1
            mb_out_content.append(line.strip(' ').strip('\t').strip('\n').strip(','))
    linenumber = int((linenumber - 12)/2+3)
    mb_useful = mb_out_content[-linenumber:-1]
    mb_id = mb_useful[:-2]
    mb_tree = mb_useful[-1]
    mb_id_dict = {}
    tree_pth = tree + ".nw"
    for i in mb_id:
        i = i.split("\t")
        mb_id_dict[i[0]]=i[1]
    with open(tree_pth,'w') as f:
        for (k,v) in mb_id_dict.items(): mb_tree = mb_tree.replace('{}[&prob='.format(k),'{}[&prob='.format(v))
        f.write(mb_tree.replace('tree con_50_majrule = [&U]',''))
    Tree = Phylo.read(tree_pth,'newick')
    for i in Tree.get_terminals(): i.comment = None
    for i in Tree.get_nonterminals(): i.comment = None
    Phylo.write(Tree,tree_pth,'newick')
    return tree_pth

def relabeltree_ete3(tree,gsmap,sptree,threshold,ismb):
    """
    gsmap is already dict
    """
    if ismb: tree = mb2nw(tree)
    genetree = Phylo.read(tree,'newick')
    if not genetree.rooted: genetree.root_at_midpoint()
    if threshold is not None: genetree = getbssubtree(genetree,threshold)
    genetree = prunepolytomy(genetree)
    safeID_origID = {}
    y = lambda x:x.strip()
    for indice,clade in enumerate(genetree.get_terminals()):
        if clade.name not in gsmap:
            logging.error("{} is not in the gsmap!".format(clade.name))
            exit(1)
        tmp = gsmap[clade.name]+'_'+str(indice)
        safeID_origID[tmp] = clade.name
        clade.name = tmp
        clade.comment = None
    Phylo.write(genetree,tree+'.labelled','newick')
    return safeID_origID,genetree,tree+'.labelled'

def recon_ete3(tree,gsmap,sptree,threshold,Nodemap,ismb,outdir):
    safeID_origID, Tree, tree_fn = relabeltree_ete3(tree,gsmap,sptree,threshold,ismb)
    y = lambda x: "_".join(x.split("_")[:-1])
    for clade in Tree.get_terminals():
        clade.comment = "&&NHX:S={}:D=N".format(y(clade.name))
    for clade in Tree.get_nonterminals():
        tips = set([y(i.name) for i in clade.get_terminals()])
        candidatenodes = {n:len(s) for n,s in Nodemap.items() if (tips-s) == (s-tips)}
        if len(candidatenodes) == 0:
            candidatenodes = {n:len(s) for n,s in Nodemap.items() if len(s-tips)>0 and len(s) > len(tips)}
        n,s = sorted(candidatenodes.items(),key=lambda x:x[1])[0]
        clade.comment = "&&NHX:S={}".format(n)
    with open(tree_fn,'r') as f: gt_content = f.read().strip("\n").strip().replace(":0.00000","")
    with open(sptree,'r') as f: st_content = f.read().strip("\n").strip()
    genetree = PhyloTree(gt_content)
    sptree = PhyloTree(st_content)
    recon_tree, events = genetree.reconcile(sptree)
    Num_S,Num_D,S_Node,D_Node=0,0,[],[]
    for event in events:
        if event.etype == 'S':
            mrca = Tree.common_ancestor(*[event.inparalogs+event.orthologs])
            S_Node.append(mrca.comment.replace("&&NHX:S=","").replace(":D=N","").replace(":D=Y",""))
            mrca.comment = mrca.comment+":D=N"
            Num_S += 1
        elif event.etype == 'D':
            mrca = Tree.common_ancestor(*[event.inparalogs+event.outparalogs])
            D_Node.append(mrca.comment.replace("&&NHX:S=","").replace(":D=N","").replace(":D=Y",""))
            mrca.comment = mrca.comment+":D=Y"
            Num_D += 1
    Phylo.write(Tree,tree+'.nhx','newick')
    #with open(tree+'_clean.nhx','w') as f:
    #    with open(tree+'.nhx') as ff:
    #        cont = ff.read()
    #        f.write(cont.replace(":0.00000",""))
    for clade in Tree.get_terminals(): clade.name = safeID_origID[clade.name]
    agora_dir = _mkdir(os.path.join(outdir,"AGORA_PREP"))
    yy = lambda x:os.path.join(agora_dir,os.path.basename(x))
    Phylo.write(Tree,yy(tree)+'_originalID.nhx','newick')
    #with open(tree+'_clean_originalID.nhx','w') as f:
    #    with open(tree+'_originalID.nhx') as ff:
    #        cont = ff.read()
    #        f.write(cont.replace(":0.00000",""))
    logging.info("In total {} speciation and {} duplication events were reconciled".format(Num_S,Num_D))
    y = lambda x:[i.replace(":D=N","").replace(":D=Y","") for i in x]
    if len(S_Node)!= 0: logging.info("Speciation event at Node {}".format(", ".join(y(S_Node))))
    if len(D_Node)!= 0: logging.info("Duplication event at Node {}".format(", ".join(y(D_Node))))

def checknode(WGDnodes,genetree,gsmap):
    gt = Phylo.read(genetree,'newick')
    for i in gt.get_terminals(): i.name = gsmap[i.name]
    mrca = gt.common_ancestor(WGDnodes)
    if len(mrca.get_terminals()) > 2:
        return False
    else:
        return True

def checkwholetree(sptree,genetree):
    with open(sptree,'r') as f: sp = f.read().strip()
    with open(genetree,'r') as f: gt = f.read().strip()
    tree1 = PhyloTree(sp)
    tree2 = PhyloTree(gt)
    rf_distance = tree1.robinson_foulds(tree2)
    print(rf_distance[0])

class Config_Hauler:
    """
    Fetch the info in the config file and implement corresponding analysis
    """
    def __init__(self,config):
        logging.info("Fetching infomation from config file")
        self.para = fetcher(config)
        self.outdir = self.para['Outdir:']
        self.OG = pd.read_csv(self.para['Orthogroup path:'],header=0,index_col=0,sep='\t')
        self.threads,self.seqtype = int(self.para['Number of threads:']),self.para['Sequences type:']
        self.data = self.para['Sequences directory:']
        self.sptree = self.para['Species tree:']
        self.wgdcheck = self.para['Check WGDnode:']
        self.ismb = True if self.para['Tree algorithm:'] == 'mrbayes' else False
        reporter(self.para)
        _mkdir(self.outdir)
        if self.para['Sequences form:'] == 'species':
            self.read_seq()
            logging.info("Writing sequences per family")
            self.write_famseq()
        else:
            self.getfam()
        logging.info("Aligning each family")
        self.aligning()
        logging.info("Inferring gene tree per family")
        self.genetree()
        if self.wgdcheck == 'yes':
            self.checkWGDnode()
        if self.para['Reconciliation:'] == 'yes':
            self.reconciliation()

    def read_seq(self):
        seq_paths = listdir(self.data)
        self.gsmap,self.SEQ = {},{}
        for seq in seq_paths:
            self.SEQ[os.path.basename(seq)] = {}
            for record in tqdm(SeqIO.parse(seq, 'fasta'),desc="Reading {}".format(os.path.basename(seq)),unit=" sequences"):
                if not (self.gsmap.get(record.id) is None):
                    logging.error("Duplicated gene id found in {}".format(os.path.basename(seq)))
                    exit(1)
                else:
                    self.gsmap[record.id] = os.path.basename(seq)
                    self.SEQ[os.path.basename(seq)][record.id] = record

    def write_famseq(self):
        OGSEQ_dir = _mkdir(os.path.join(self.outdir,"OG_SEQ"))
        parent = os.getcwd()
        os.chdir(OGSEQ_dir)
        self.Fam_list = Fam_list = list(self.OG.index)
        self.Fam_path = {i:os.path.join(OGSEQ_dir,i+'.'+self.seqtype) for i in Fam_list}
        self.SOGs = list(self.OG.index[[find_singleton(self.OG.loc[fam,:]) for fam in Fam_list]])
        self.NonSOGs = list(set(Fam_list)-set(self.SOGs))
        if len(set(self.OG.index)) != len(Fam_list):
            element_freq = Counter(Fam_list)
            duplicated = [element for element,freq in element_freq.items() if freq != 1]
            logging.error("Duplicated gene family id found for {}".format(", ".join(duplicated)))
            exit(1)
        Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(write_seq)(Fam_list[i],self.OG.loc[Fam_list[i],:],self.SEQ,self.seqtype) for i in trange(len(Fam_list)))
        os.chdir(parent)
        if self.para['Translation:'] == 'yes':
            logging.info("Writing translated sequences per family")
            OGSEQPEP_dir = _mkdir(os.path.join(self.outdir,"OG_SEQ_TRANSLATE"))
            os.chdir(OGSEQPEP_dir)
            to_stop,cds = self.para['Translation parameters:'].split(',') 
            Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(write_seq_translate)(Fam_list[i],self.OG.loc[Fam_list[i],:],self.SEQ,to_stop,cds) for i in trange(len(Fam_list)))
            self.Fam_path = {i:os.path.join(OGSEQPEP_dir,i+'.pep') for i in Fam_list}
            os.chdir(parent)

    def getfam(self):
        self.Fam_list = Fam_list = list(self.OG.index)
        self.SOGs = list(self.OG.index[[find_singleton(self.OG.loc[fam,:]) for fam in Fam_list]])
        self.NonSOGs = list(set(Fam_list)-set(self.SOGs))
        self.Fam_path = {fam:path for fam,path in zip(sorted(Fam_list),listdir(self.data,align=True))}
        if self.para['Translation:'] == 'yes' and self.para['Sequences type:'] == 'nucleotide':
            logging.info("Writing translated sequences per family")
            OGSEQPEP_dir = _mkdir(os.path.join(self.outdir,"OG_SEQ_TRANSLATE"))
            to_stop,cds = self.para['Translation parameters:'].split(',')
            Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(write_seq_translate_fammode)(Fam_list[i],self.Fam_path[Fam_list[i]],to_stop,cds,OGSEQPEP_dir) for i in trange(len(Fam_list)))
            self.Fam_path = {i:os.path.join(OGSEQPEP_dir,i+'.pep') for i in Fam_list}

    def aligning(self):
        if self.para['Aligner parameters:'] == 'default': options = ''
        else:
            y = lambda x:[i.strip() for i in x]
            options = deal_options(y(self.para['Aligner parameters:'].split(',')))
        self.OGALIGN_dir = OGALIGN_dir = _mkdir(os.path.join(self.outdir,"OG_ALIGNMENT_TREE"))
        if self.para['Aligner:'] == 'mafft':
            z = lambda x:os.path.join(OGALIGN_dir,x+".mafft")
            Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(mafft)(self.Fam_path[self.NonSOGs[i]],options,z(self.NonSOGs[i])) for i in trange(len(self.NonSOGs)))
        elif self.para['Aligner:'] == 'muscle':
            z = lambda x:os.path.join(OGALIGN_dir,x+".muscle")
            Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(muscle)(self.Fam_path[self.NonSOGs[i]],options,z(self.NonSOGs[i])) for i in trange(len(self.NonSOGs)))
        elif self.para['Aligner:'] == 'prank':
            z = lambda x:os.path.join(OGALIGN_dir,x+".prank")
            Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(prank)(self.Fam_path[self.NonSOGs[i]],options,z(self.NonSOGs[i])) for i in trange(len(self.NonSOGs)))
        self.Aln_Path = {fam:path for fam,path in zip(sorted(self.NonSOGs),listdir(OGALIGN_dir,align=True))}

    def genetree(self):
        y = lambda x:[i.strip() for i in x]
        if self.para['Tree algorithm parameters:'] == 'default': options = []
        elif self.para['Tree algorithm:'] == 'mrbayes':
            options = {}
            for p in y(self.para['Tree algorithm parameters:'].split(',')):
                if options.get(p.split(' ')[0]) is None:
                    options[p.split(' ')[0]] = p.split(' ')[1:]
                else:
                    options[p.split(' ')[0]] = options[p.split(' ')[0]]+p.split(' ')[1:]
        else:
            options = deal_options(y(self.para['Tree algorithm parameters:'].split(',')))
        keys = sorted(self.Aln_Path.keys())
        values = {i:self.Aln_Path[j] for i,j in enumerate(keys)}
        if self.para['Tree algorithm:'] == 'iqtree':
            treepaths = Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(iqtree)(values[i],options) for i in trange(len(keys)))
        if self.para['Tree algorithm:'] == 'iqtree2':
            treepaths = Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(iqtree2)(values[i],options) for i in trange(len(keys)))
        if self.para['Tree algorithm:'] == 'fasttree':
            treepaths = Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(fasttree)(values[i],options) for i in trange(len(keys)))
        if self.para['Tree algorithm:'] == 'mrbayes':
            parent = os.getcwd()
            os.chdir(self.OGALIGN_dir)
            treepaths = Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(mrbayes)(os.path.basename(values[i]),options) for i in trange(len(keys)))
            os.chdir(parent)
        self.treepaths = [i for i in treepaths if not i is None]

    def reconciliation(self):
        if not os.path.exists(self.sptree):
            logging.error("Please provide a species tree for reconciliation!")
            exit(1)
        df = pd.read_csv(labelinternalnodes(self.sptree),header=None,index_col=None,sep='\t')
        if self.para['Branch supportness filter:'] == '0': self.threshold = None
        elif float(self.para['Branch supportness filter:']) > 1: self.threshold = float(self.para['Branch supportness filter:'])/100
        else: self.threshold = float(self.para['Branch supportness filter:'])
        self.Nodemap = {n:set(s.split(", ")) for n,s in zip(df[0],df[1])}
        if self.para['Reconciler:'] == 'ete3':
            Parallel(n_jobs=self.threads,backend='multiprocessing')(delayed(recon_ete3)(self.treepaths[i],self.gsmap,self.sptree,self.threshold,self.Nodemap,self.ismb,self.outdir) for i in trange(len(self.treepaths)))

    def checkWGDnode(self):
        sptree = Phylo.read(self.sptree,'newick')
        WGD_nodes,GoodFam,BadFam = [],[],[]
        for clade in sptree.get_terminals():
            if clade.name.endswith('_ap1') or clade.name.endswith('_ap2'): WGD_nodes.append(clade.name)
        for i in self.treepaths:
            if self.ismb: i = mb2nw(i)
            #checkwholetree(self.para['Species tree:'],i)
            if checknode(WGD_nodes,i,self.gsmap): GoodFam.append(i)
            else: BadFam.append(i)
        logging.info("In total {} families passed the topology test while {} failed".format(len(GoodFam),len(BadFam)))
        Passed_path = os.path.join(self.outdir,os.path.basename(self.para['Orthogroup path:'])+'.Passed')
        Failed_path = os.path.join(self.outdir,os.path.basename(self.para['Orthogroup path:'])+'.Failed')
        y = lambda x: os.path.basename(x).replace('.treefile','').replace('.FastTree','').replace('.nexus.con.tre.nw','').replace('.muscle','').replace('.mafft','').replace('.prank','').replace('.nexus.con.tre','')
        self.OG.loc[[y(i) for i in GoodFam]].to_csv(Passed_path,header=True,index=True,sep='\t')
        self.OG.loc[[y(i) for i in BadFam]].to_csv(Failed_path,header=True,index=True,sep='\t')





# TODO: still has to be updated
import numpy as np
import logging
from sklearn import mixture
import scipy.stats as ss
import plumbum as pb
import matplotlib

if not 'DISPLAY' in pb.local.env:
    matplotlib.use('Agg')  # use this backend when no X server
import matplotlib.pyplot as plt
import seaborn as sns


def filter_group_data(
        df, aln_len=300, min_ks=0, max_ks=5,
        weights_outliers_included=False
):
    """
    Filter a Ks data frame for modling purposes.

    :param df: data frame
    :param aln_id: alignment identity threshold
    :param aln_len: alignment length threshold
    :param aln_cov: alignment coverage threshold
    :param min_ks: minimum Ks value
    :param max_ks: maximum Ks value
    :param weights_outliers_included: boolean, adapt weights when removing
        outliers
    :return: filtered data frame
    """
    # pre node-grouping filters, irrespective of outlier removal
    df = df.dropna()
    #df = df[df["alnlen"] >= aln_len]
    df = df[df['alignmentlength']>= aln_len]

    # Ks range filters
    # if one filters before the node-weighting, the weights are adapted with
    # respect to the outlier filtering (as in Vanneste et al. 2013)
    if not weights_outliers_included:
        df = df[df["dS"] > min_ks]
        df = df[df["dS"] <= max_ks]

    # grouping
    df = df.groupby(['family', 'node']).mean()

    if weights_outliers_included:
        df = df[df["dS"] > min_ks]
        df = df[df["dS"] <= max_ks]

    return df


def get_array_for_mixture(df):
    """
    Get an array of log transformed Ks values.

    :param df: data frame
    :return: array
    """
    X = np.array(df["dS"].dropna())
    X = X[X > 0]
    X = np.log(X).reshape(-1, 1)
    return X


def reflect(data):
    """
    Reflect an array around it's left boundary

    :param data: np.array
    :return: np.array
    """
    reflection = -1 * (data - np.min(data)) + np.min(data)
    return np.hstack([data, reflection])


def reflected_kde(df, min_ks, max_ks, bandwidth, bins, out_file):
    """
    Perform Kernel density estimation (KDE) with reflected data.
    The data frame is assumed to be grouped already by 'Node'.

    :param df: data frame
    :param min_ks: minimum Ks value (best is to use 0, for reflection purposes)
    :param max_ks: maximum Ks value
    :param bandwidth: bandwidth (None results in Scott's rule of thumb)
    :param bins: number of histogram bins
    :param out_file: output file
    :return: nada
    """
    ks = np.array(df['dS'])
    ks_reflected = reflect(ks)
    fig, ax = plt.subplots(figsize=(9, 4))
    if bandwidth:
        ax = sns.distplot(
                ks_reflected, bins=bins * 2, ax=ax,
                hist_kws={"rwidth": 0.8, "color": "k", "alpha": 0.2},
                kde_kws={"bw": bandwidth},
                color="k"
        )
    else:
        ax = sns.distplot(
                ks_reflected, bins=bins * 2, ax=ax,
                hist_kws={"rwidth": 0.8, "color": "k", "alpha": 0.2},
                color="k"
        )
    ax.set_xlim(min_ks, max_ks)
    sns.despine(offset=5, trim=False)
    ax.set_ylabel("Density")
    ax.set_xlabel("$K_{\mathrm{S}}$")
    fig.savefig(out_file, bbox_inches='tight')


def fit_gmm(X, n1, n2, max_iter=100, n_init=1, **kwargs):
    """
    Compute Gaussian mixtures for different numbers of components

    :param X: data frame (log transformed Ks values)
    :param n1: minimum number of components
    :param n2: maximum number of components
    :param max_iter: maximum number of iterations
    :param n_init: number of k-means initializations
    :param kwargs: other keyword args for `GaussianMixture`
    :return: models, bic, aic, best model
    """
    # fit models with 1 to n components
    N = np.arange(n1, n2 + 1)
    models = [None for i in range(len(N))]
    for i in range(len(N)):
        logging.info("Fitting GMM with {} components".format(N[i]))
        models[i] = mixture.GaussianMixture(
                n_components=N[i], covariance_type='full', max_iter=max_iter,
                n_init=n_init, **kwargs
        ).fit(X)
        logging.info("Component mean, variance, weight: ")
        log_components(models[i])

    # compute the AIC and the BIC
    aic = [m.aic(X) for m in models]
    bic = [m.bic(X) for m in models]
    best = models[np.argmin(bic)]

    return models, bic, aic, best


def log_components(m):
    for j in range(len(m.means_)):
        logging.info(".. {0:.3f}, {1:.3f}, {2:.3f}".format(
                np.exp(m.means_[j][0]),
                m.covariances_[j][0][0], m.weights_[j])
        )


def fit_bgmm(X, n1, n2, gamma=1e-3, max_iter=100, n_init=1, **kwargs):
    """
    Compute Bayesian Gaussian mixture

    :param X: data frame (log transformed Ks values)
    :param n1: minimum number of components
    :param n2: maximum number of components
    :param gamma: inverse of regularization strength
    :param max_iter: maximum number of iterations
    :param n_init: number of k-means initializations
    :param kwargs: other keyword args for `GaussianMixture`
    :return: models
    """
    # fit models with 1 to n components
    N = np.arange(n1, n2 + 1)
    models = [None for i in range(len(N))]

    for i in range(len(N)):
        logging.info("Fitting BGMM with {} components".format(N[i]))
        models[i] = mixture.BayesianGaussianMixture(
                weight_concentration_prior=gamma, n_init=n_init,
                n_components=N[i], covariance_type='full', max_iter=max_iter,
                **kwargs
        ).fit(X)
        log_components(models[i])

    return models


def plot_mixture(model, data, ax, l=0, u=5, color='black', alpha=0.2,
                 log=False, bins=25, alpha_l1=1):
    """
    Plot a mixture model. Assumes a log-transformed model and data
    and will back-transform.

    Note from scipy docs:
    ---------------------
    A common parametrization for a lognormal random variable Y
    is in terms of the mean, mu, and standard deviation, sigma,
    of the unique normally distributed random variable X such
    that exp(X) = Y. This parametrization corresponds to setting
    s = sigma and scale = exp(mu).

    So since we fit a normal mixture on the logscale, we can get
    the lognormal by setting scale to np.exp(mean[k]) and s to
    np.sqrt(variance[k]) for component k.

    :param model: model object
    :param data: data array
    :param ax: figure ax
    :param l: lower Ks limit
    :param u: upper Ks limit
    :param color: color for histogram
    :param alpha: alpha value
    :param log: plot on log scale?
    :param bins: number of histogram bins
    :param alpha_l1: alpha value for mixture lines
    :return: ax
    """
    x = np.linspace(l, u, 1000).reshape((-1, 1))
    if not log:
        data = np.exp(data)
    data = data[data >= l]
    data = data[data <= u]
    bins = ax.hist(
            data, bins, rwidth=0.8, color=color, alpha=alpha, density=True)
    maxy = max(bins[0])
    means = model.means_
    varcs = model.covariances_
    weights = model.weights_
    mix = None
    first = True
    for k in range(len(means)):
        if not log:
            curve = ss.lognorm.pdf(
                    x, scale=np.exp(means[k]), s=np.sqrt(varcs[k])) * weights[k]
        else:
            curve = ss.norm.pdf(
                    x, loc=means[k], scale=np.sqrt(varcs[k])) * weights[k]
        ax.plot(x, curve, '--k', color='black', alpha=0.4)
        if first:
            mix = curve
            first = False
        else:
            mix += curve
    ax.plot(x, mix, color='black', alpha=alpha_l1)
    ax.set_xlim(l, u)
    if log:
        ax.set_xlabel("$\mathrm{log}(K_{\mathrm{S}})$")
    else:
        ax.set_xlabel("$K_{\mathrm{S}}$")
    ax.set_ylabel("Density")
    return ax


def inspect_aic(aic):
    """
    Evaluate the Akaike information criterion results for mixture models.

    :param aic: AIC values
    """
    im = np.argmin(aic)
    logging.info("")
    logging.info("AIC assessment:")
    logging.info("min(AIC) = {:.2f} for model {}".format(aic[im], im + 1))
    logging.info("Relative probabilities compared to model {}:".format(im + 1))
    logging.info("   /                          \\")
    logging.info("   |      (min(AIC) - AICi)/2 |")
    logging.info("   | p = e                    |")
    logging.info("   \                          /")
    for i, aic_i in enumerate(aic):
        p_i = np.exp((aic[im] - aic_i) / 2)
        logging.info(".. model{:>4}: p = {:.4f}".format(i + 1, p_i))
    logging.info("")


def inspect_bic(bic):
    """
    Evaluate the BIC values

    :param bic: BIC values
    """
    im = np.argmin(bic)
    l = [
        "0 to  2:   Very weak",
        "2 to  6:    Positive",
        "6 to 10:      Strong",
        "    >10: Very Strong"
    ]
    logging.info("")
    logging.info("Delta BIC assessment: ")
    logging.info("min(BIC) = {:.2f} for model {}".format(bic[im], im + 1))
    for i, bic_i in enumerate(bic):
        dbic = bic_i - bic[im]
        j = 0
        if dbic > 2: j = 1
        if dbic > 6: j = 2
        if dbic > 10: j = 3
        logging.info(".. model{:>4}: delta(BIC) = {:>8.2f} ({})".format(
                i + 1, dbic, l[j]))
    logging.info("")


def plot_probs(m, ax, l=0.0, u=5, ylab=True):
    """
    Plot posterior component probabilities

    :param m: model
    :param ax: figure ax
    :param l: lower Ks limit
    :param u: upper Ks limit
    :param ylab: plot ylabel?
    :return: ax
    """
    if l == 0:
        l = 0.005
    x = np.linspace(l, u, 1000).reshape(-1, 1)
    p = m.predict_proba(np.log(x))
    p = p.cumsum(1).T
    x = np.linspace(l, u, 1000)
    order = tuple(np.argsort(m.means_.reshape((1, -1)))[0])
    alphas = np.linspace(0.2, 1, p.shape[0])[order,]
    for i, array in enumerate(p):
        if i == 0:
            ax.fill_between(
                    x, 0, p[i], color='gray', alpha=alphas[i],
                    label=order[i] + 1)
        elif i == len(alphas) - 1:
            ax.fill_between(
                    x, p[i - 1], 1, color='gray', alpha=alphas[i],
                    label=order[i] + 1)
        else:
            ax.fill_between(
                    x, p[i - 1], p[i], color='gray', alpha=alphas[i],
                    label=order[i] + 1
            )
    ax.set_xlim(0, u)
    ax.set_ylim(0, 1)
    ax.set_xlabel('$K_{\mathrm{S}}$')
    if ylab:
        ax.set_ylabel('$P(class|K_{\mathrm{S}})$')
    ax.legend(frameon=True)
    return ax


def plot_aic_bic(aic, bic, min_n, max_n, out_file):
    """
    Plot AIC and BIC curves

    :param aic: aic values
    :param bic: bic values
    :param out_file: output file
    :return: nada
    """
    x_range = list(range(min_n, max_n + 1))
    fig, axes = plt.subplots(1, 2, figsize=(12, 3))
    axes[0].plot(np.arange(1, len(aic) + 1), aic, color='k', marker='o')
    axes[0].set_xticks(list(range(1, len(aic) + 1)))
    axes[0].set_xticklabels(x_range)
    axes[0].grid(ls=":")
    axes[0].set_ylabel("AIC")
    axes[0].set_xlabel("# components")
    axes[1].plot(np.arange(1, len(bic) + 1), bic, color='k', marker='o')
    axes[1].set_xticks(list(range(1, len(bic) + 1)))
    axes[1].set_xticklabels(x_range)
    axes[1].grid(ls=":")
    axes[1].set_ylabel("BIC")
    axes[1].set_xlabel("# components")
    fig.tight_layout()
    fig.savefig(out_file)


def plot_bars_weights(model, ax):
    """
    Plot Component weights

    :param model: model
    :param ax: figure ax
    :return: ax
    """
    ax.bar(np.arange(1, model.n_components + 1), model.weights_, color='k',
           alpha=0.2)
    for i, mn in enumerate(model.means_):
        ax.text(
                x=i + 1, y=model.weights_[i] + 0.1,
                horizontalalignment='center',
                s='$\hat{{\mu}} = {:.2f}$'.format(np.exp(mn)[0])
        )
    ax.set_ylabel("weight")
    ax.set_xlabel("component")
    ax.set_ylim(0, 1)
    return ax


def plot_all_models_gmm(models, data, l, u, bins, out_file):
    """
    Plot a bunch of GMMs.

    :param models: list of GMM model objects
    :param data: Ks array
    :param l: lower Ks limit
    :param u: upper Ks limit
    :param bins: number of histogram bins
    :param out_file: output file
    :return: nada
    """
    fig, axes = plt.subplots(len(models), 3, figsize=(15, 3 * len(models)))
    for i, model in enumerate(models):
        plot_mixture(model, data, axes[i, 0], l, u, bins=bins)
        plot_mixture(model, data, axes[i, 1], log=True, l=np.log(l + 0.0001),
                     u=np.log(u), bins=bins)
        plot_probs(model, axes[i, 2], l, u)
    sns.despine(offset=5)
    fig.tight_layout()
    fig.savefig(out_file)


def plot_all_models_bgmm(models, data, l, u, bins, out_file):
    """
    Plot a bunch of BGMMs.

    :param models: list of GMM model objects
    :param data: Ks array
    :param l: lower Ks limit
    :param u: upper Ks limit
    :param bins: number of histogram bins
    :param out_file: output file
    :return: nada
    """
    fig, axes = plt.subplots(len(models), 4, figsize=(20, 3 * len(models)))
    for i, model in enumerate(models):
        plot_mixture(model, data, axes[i, 0], l, u, bins=bins)
        plot_mixture(model, data, axes[i, 1], log=True, l=np.log(l + 0.0001),
                     u=np.log(u), bins=bins)
        plot_probs(model, axes[i, 2], l, u)
        plot_bars_weights(model, axes[i, 3])
    sns.despine(offset=5)
    fig.tight_layout()
    fig.savefig(out_file)


def get_component_probabilities(df, model):
    """
    Get paralogs from a mixture model component.

    :param df: Ks distribution pandas data frame
    :param model: mixture model
    :return: data frame
    """
    df = df.dropna()
    df = df.drop_duplicates(keep='first')
    df['log(dS)'] = np.log(df['dS'])
    df = df.replace([np.inf, -np.inf], np.nan)
    df = df.dropna()

    p = model.predict_proba(np.array(df['log(dS)']).reshape(-1, 1))
    order = np.argsort([x[0] for x in model.means_])
    order_dict = {i: order[i] for i in range(len(order))}
    for c in range(len(order)):
        col = 'p_component{}'.format(order_dict[c] + 1)
        df[col] = p[:, c]

    return df



import pandas as pd
import click
import itertools
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats
import seaborn as sns
from rich.logging import RichHandler
import logging

def kde_mode(kde_x, kde_y):
    maxy_iloc = np.argmax(kde_y)
    mode = kde_x[maxy_iloc]
    return mode, max(kde_y)

def get_totalH(Hs):
    CHF = 0
    for i in Hs: CHF = CHF + i
    return CHF

def calculateHPD(train_in,per):
    sorted_in = np.sort(train_in)
    upper_bound = np.percentile(train_in, per)
    lower_bound = np.percentile(train_in, 100-per)
    upper_bound_indice,lower_bound_indice = 0,0
    cutoff,candidates = int(np.ceil(per*len(sorted_in)/100)),[]
    for i,v in enumerate(sorted_in):
        if v >= upper_bound:
            upper_bound_indice = i
            break
    for i,v in enumerate(sorted_in):
        if v >= lower_bound:
            lower_bound_indice = i
            break
    for (x,y) in itertools.product(np.arange(0,lower_bound_indice,1,dtype=int), np.arange(upper_bound_indice,len(sorted_in),1,dtype=int)):
        if (y-x+1) >= cutoff: candidates.append((sorted_in[y] - sorted_in[x],(x,y)))
    lower,upper = sorted(candidates, key=lambda y: y[0])[0][1][0],sorted(candidates, key=lambda y: y[0])[0][1][1]
    return sorted_in[upper],sorted_in[lower]

@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.option('--verbosity', '-v', type=click.Choice(['info', 'debug']), default='info', help="Verbosity level, default = info.")
def cli(verbosity):
    """
    This script is for post WGD dating visualization
    """
    logging.basicConfig(
        format='%(message)s',
        handlers=[RichHandler()],
        datefmt='%H:%M:%S',
        level=verbosity.upper())
    logging.info("Proper Initiation")
    pass

@cli.command(context_settings={'help_option_names': ['-h', '--help']})
@click.argument('posterior', type=click.Path(exists=True))
@click.option('--percentile','-p',show_default=True, default=90,help = "percentile of posteriors")
@click.option('--indexcol', is_flag=True,help="first column as index col")
@click.option('--hpd', is_flag=True,help="calculate HPD CI instead of Equal-tail CI")
@click.option('--title','-t',show_default=True, default="WGD date",help = "title of plot")
@click.option('--output','-o',show_default=True, default="xx_dates_inM.pdf",help = "filename of output")
def postdis(posterior,percentile,indexcol,hpd,title,output):
    """
    Plot the posterior distribution
    """
    if indexcol: df = pd.read_csv(posterior,header=0,index_col=0,sep='\t').dropna()
    else: df = pd.read_csv(posterior,header=0,index_col=None,sep='\t').dropna()
    fig, ax = plt.subplots()
    train_in = np.array(df.loc[:,df.columns[0]])*100
    upper = np.percentile(train_in, percentile+(100-percentile)/2)
    lower = np.percentile(train_in, (100-percentile)/2)
    maxm,minm,mean,median = float(train_in.max()),float(train_in.min()),float(train_in.mean()),np.median(train_in)
    ax.set_xlim(minm-10,maxm+10)
    kde_x = np.linspace(minm,maxm,num=500)
    kde_y=stats.gaussian_kde(train_in,bw_method='silverman').pdf(kde_x)
    mode, maxim = kde_mode(kde_x, kde_y)
    Hs, Bins, patches = ax.hist(train_in,bins = np.linspace(0, 200, num=201),color='g', alpha=0.8, rwidth=0.8,label='Raw dates')
    CHF = get_totalH(Hs)
    scaling = CHF*1
    ax.plot(kde_x, kde_y*scaling, color='k',alpha=0.8, ls = '-',lw = 1)
    if hpd:
        upper_HPD,lower_HPD = calculateHPD(train_in,percentile)
        plt.axvline(x = upper_HPD, color = 'k', alpha = 0.8, ls = '-.', lw = 1,label='{}% HPD CI Upper {:.2f}'.format(percentile,upper_HPD))
        plt.axvline(x = lower_HPD, color = 'k', alpha = 0.8, ls = '-.', lw = 1,label='{}% HPD CI Lower {:.2f}'.format(percentile,lower_HPD))
    else:
        plt.axvline(x = upper, color = 'k', alpha = 0.8, ls = '-.', lw = 1,label='{}% Equal-tail CI Upper {:.2f}'.format(percentile,upper))
        plt.axvline(x = lower, color = 'k', alpha = 0.8, ls = '-.', lw = 1,label='{}% Equal-tail CI Lower {:.2f}'.format(percentile,lower))
    plt.axvline(x = mean, color = 'k', alpha = 0.8, ls = ':', lw = 1,label='Posterior Mean {:.2f}'.format(mean))
    plt.axvline(x = median, color = 'k', alpha = 0.8, ls = '--', lw = 1,label='Median {:.2f}'.format(median))
    plt.axvline(x = mode, color = 'k', alpha = 0.8, ls = '-', lw = 1,label='Mode {:.2f}'.format(mode))
    ax.legend(loc=0,fontsize='small',frameon=False)
    ax.set_xlabel("Million years ago")
    ax.set_ylabel("Counts")
    sns.despine(offset=1)
    plt.title(title)
    fig.tight_layout()
    if output != "xx_dates_inM.pdf": fig.savefig(output)
    else: fig.savefig(posterior+'_dates_inM.pdf')
    plt.close()

if __name__ == '__main__':
	cli()



#!/usr/bin/python3.8
# contact: heche@psb.vib-ugent.be
# https://stackoverflow.com/questions/43658870/requirements-txt-vs-setup-py

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='wgd',
    version='2.0.34',
    packages=['wgd'],
    url='http://github.com/heche-psb/wgd',
    license='GPL',
    author='Hengchi Chen',
    author_email='heche@psb.vib-ugent.be',
    description='wgd',
    long_description=long_description,
    long_description_content_type="text/markdown",
    py_modules=['cli'],
    include_package_data=True,
    install_requires=[
       'attrs==20.3.0',
       'biopython==1.76',
       'click==7.1.2',
       'colorama==0.4.4',
       'commonmark==0.9.1',
       'cycler==0.10.0',
       'ete3>=3.1.3',
       'fastcluster==1.1.28',
       'humanfriendly==8.2',
       'importlib-metadata==3.1.0',
       'iniconfig==1.1.1',
       'Jinja2==2.11.2',
       'joblib==0.11',
       'KDEpy==1.1.0',
       'kiwisolver==1.2.0',
       'MarkupSafe==1.1.1',
       'matplotlib==3.2.2',
       'numpy>=1.19.0,!=1.24.3',
       'numexpr>=2.7.3',
       'packaging==20.4',
       'pandas<=1.4.4',
       'Pillow<=8.4.0',
       'pluggy==0.13.1',
       'plumbum==1.6.9',
       'progressbar2==3.51.4',
       'py==1.9.0',
       'Pygments==2.7.2',
       'pyqtwebengine<5.13',
       'pyqt5<5.13',
       'pyparsing==2.4.7',
       'pytest==6.1.2',
       'python-dateutil>=2.8.2',
       'python-utils==2.4.0',
       'pytz==2020.1',
       'PyYAML==5.3.1',
       'rich==12.5.1',
       'scikit-learn==0.23.1',
       'scikit-learn-extra==0.2.0',
       'scipy<=1.5.4',
       'seaborn==0.10.1',
       'six==1.15.0',
       'threadpoolctl==2.1.0',
       'toml==0.10.2',
       'tornado==6.0.4',
       'tqdm>=4.64.1',
       'typing-extensions>=3.10.0.0',
       'zipp==3.4.0'
    ],
    entry_points='''
        [console_scripts]
        wgd=cli:cli
    ''',
)
